# -*- encoding: utf-8 -*-

'''
Created on 6 июл. 2021 г.

@author: dvd
'''
from __future__ import unicode_literals

from rest_framework import serializers
from termidesk.api.fields import TypedSerializerMethodField
from termidesk.core.managers.PolicyManager import PolicyManager
from termidesk.policies.DetachableUserProfileEnabledPolicy import DetachableUserProfileEnabledPolicy


class ServiceOwnerDetail(serializers.Serializer):
    '''
    Read-only DRF сериалайзер для подготовки данных о владельце рабочего места (сервиса)
    '''

    username = serializers.SerializerMethodField(help_text="Имя пользователя")
    auth_id = TypedSerializerMethodField(
        field_type=serializers.IntegerField, help_text="ID аутентификатора (PK) пользователя")
    auth_uuid = TypedSerializerMethodField(
        field_type=serializers.UUIDField, help_text="UUID аутентификатора пользователя")
    auth_smallname = serializers.SerializerMethodField(help_text="Короткое имя аутентификатора")
    auth_realm = serializers.SerializerMethodField(help_text="Домен аутентификатора")

    def _get_auth_realm(self, authenticator):
        """
        :param instance: экземпляр одного из дочерних классов класса :py:class:`uds.core.Auth.Authenticator`
        """
        # Получаем значение realm аутентификатора
        # Пока действительно только для ALD и FreeIPA аутентификаторов
        realm_field = getattr(authenticator, 'realm', {'value': ''})
        realm_value = getattr(realm_field, 'value', '')
        return realm_value

    def _get_db_auth(self, instance):
        """
        Возвращает экземпляр модели аутентификатора.
        :param instance: экземпляр класса :py:class:`uds.models.User`
        """
        authenticator = instance.getManager()
        return authenticator.dbAuthenticator()

    def get_username(self, instance):
        """ :param instance: экземпляр класса :py:class:`uds.models.User` """
        return instance.name

    def get_auth_id(self, instance):
        """ :param instance: экземпляр класса :py:class:`uds.models.User` """
        db_auth = self._get_db_auth(instance)
        return db_auth.pk

    def get_auth_uuid(self, instance):
        """ :param instance: экземпляр класса :py:class:`uds.models.User` """
        db_auth = self._get_db_auth(instance)
        return db_auth.uuid

    def get_auth_smallname(self, instance):
        """ :param instance: экземпляр класса :py:class:`uds.models.User` """
        db_auth = self._get_db_auth(instance)
        return db_auth.small_name

    def get_auth_realm(self, instance):
        """ :param instance: экземпляр класса :py:class:`uds.models.User` """
        authenticator = instance.getManager()
        realm = self._get_auth_realm(authenticator)
        return realm


class ServiceInfo(serializers.Serializer):
    '''
    Read-only DRF сериалайзер для подготовки данных о рабочем месте (сервисе)
    '''

    name = serializers.CharField(source="getName", read_only=True, help_text="Название машины")
    user = ServiceOwnerDetail(read_only=True, allow_null=True, help_text="Сведения о пользователе")
    assigned = TypedSerializerMethodField(
        field_type=serializers.BooleanField, help_text="Назначение пользователю")

    def get_assigned(self, instance):
        return instance.user is not None


class ServiceSettings(serializers.Serializer):
    '''
    Сериалайзер для подготовки данных о настройках рабочего места
    '''

    debug = TypedSerializerMethodField(
        field_type=serializers.BooleanField, help_text="Режим отладки")
    detachableUserProfileEnabled = TypedSerializerMethodField(
        field_type=serializers.BooleanField, help_text="Отделяемый пользовательский профиль")

    def get_debug(self, instance):
        return instance.deployed_service.debug_mode

    def get_detachableUserProfileEnabled(self, instance):
        return PolicyManager().get_deployed_service_policy(
            DetachableUserProfileEnabledPolicy, instance.deployed_service
        ).is_allowed()
