from importlib import import_module

from rest_framework.response import Response
from rest_framework.schemas.openapi import SchemaGenerator
from rest_framework.views import APIView

from termidesk.api.permissions import IsAuthAdmin


class OpenApiSchemaView(APIView):
    schema = None  # exclude swagger path from the API documentation
    permission_classes = [IsAuthAdmin]

    def get(self, request, **kwargs):

        module = kwargs.pop('module', None)
        submodule = kwargs.pop('submodule', None)

        generator = SchemaGenerator(
            title='Документация Termidesk API {} {}'.format(module, submodule),
            url=get_module_url(module, submodule),
            urlconf=get_module_urlconf(module, submodule),
        )
        schema = generator.get_schema()

        # Добавление информации о механизмах аутентификации
        schema.setdefault('components', {})
        schema['components']['securitySchemes'] = {
            'bearerAuth': {
                'type': 'http',
                'scheme': 'bearer',
                'description': ('Для аутентифкации в модулях раздела **external**.  '
                                'Получить токен можно в модуле **auth** по url /external/login.')
            },
            'legacyTokenAuth': {
                'type': 'apiKey',
                'in': 'header',
                'name': 'X-Auth-Token',
                'description': ('Альтернативный способ аутентифкации в модулях раздела **webui**.   \n'
                                'Представляет собой идентификатор HTTP сессии пользователя.   \n'
                                'Может быть получен в модуле **auth** по url /legacy/login.')
            }
        }

        for path in schema['paths']:
            if path.startswith("/api/external"):
                for op in schema['paths'][path]:
                    schema['paths'][path][op]['security'] = [
                        {
                            'bearerAuth': []
                        },
                    ]

            if path.startswith("/api/webui"):
                for op in schema['paths'][path]:
                    schema['paths'][path][op]['security'] = [
                        {
                            'legacyTokenAuth': []
                        },
                    ]

        return Response(schema)


def get_module_urlconf(module, submodule):
    """Возвращает строку или целый модуль urlconf,
     по которому необходимо строить схему openapi
     """

    urlconf = 'termidesk.api.urls'

    if not module:
        return urlconf

    def filterer(url_pattern):
        return url_pattern.namespace == module

    module_resolver = next(filter(filterer, import_module(urlconf).urlpatterns), None)

    if not module_resolver:
        return urlconf

    if submodule and submodule in module_resolver.namespace_dict:
        return module_resolver.namespace_dict[submodule][1].urlconf_name

    return module_resolver.urlconf_name


def get_module_url(module, submodule):
    """Возвращает строку с url префиксом методов API"""

    url = '/api'
    if module:
        url += '/' + module
    if module and submodule:
        url += '/' + submodule
    return url
