'''
Created on 10 февр. 2022 г.

@author: dvd
'''
from rest_framework.schemas.openapi import AutoSchema
from uds.core.ui import gui

from .serializers.gui import GUICharFieldSerializer
from .serializers.gui import GUIBoolFieldSerializer

import logging

logger = logging.getLogger(__name__)


class GuiFieldsSchema(AutoSchema):
    serializers_classes = [
        GUICharFieldSerializer,
        GUIBoolFieldSerializer
    ]

    def get_components(self, path, method):

        components = dict()
        for clazz in self.serializers_classes:
            serializer = clazz()
            component_name = self.get_component_name(serializer)
            content = self.map_serializer(serializer)
            components[component_name] = content

        return components

    def get_operation(self, path, method):
        res = AutoSchema.get_operation(self, path, method)

        schema = res['responses']['200']['content']['application/json']['schema']

        refs = []
        for clazz in self.serializers_classes:
            serializer = clazz()
            refs.append(self._get_reference(serializer))

        schema['items'] = {
            'anyOf': refs
        }

        return res


class SendMessageSchema(AutoSchema):
    """Схема для корректной генерации документации"""
    def get_operation(self, path, method):
        responses = AutoSchema.get_operation(self, path, method)

        del(responses['responses']['201'])

        responses['responses']['200'] = {
            'content': {
                'application/json': {
                    'schema': {
                        'type': 'string',
                        'example': 'ok'
                    }
                }
            }
        }

        responses['responses']['400'] = {
            'content': {
                'application/json': {
                    'schema': {
                        'type': 'object',
                        'required': ['detail', 'status_code'],
                        'properties': {
                            'detail': {
                                'type': 'string',
                                'example': 'Длина сообщения >1000 символов'
                            },
                            'validation': {
                                'type': 'object',
                                'description': (
                                    "Содержит ошибки валидации значений атрибутов.  "
                                    "Может отсутствовать в ответе."
                                    ),
                                'properties': {
                                    "non_field_errors": {
                                        'type': 'array',
                                        'items': {
                                            'type': 'string',
                                            'nullable': 'true'
                                        }
                                    },
                                    "field_errors": {
                                        'type': 'object',
                                        "additionalProperties": {
                                            'type': 'array',
                                            'items': {
                                                'type': 'string',
                                            }
                                        },
                                        "example": {
                                            "message": [
                                                "Длина сообщения >1000 символов"
                                            ],
                                            "level": [
                                                ("Значения bad_val нет среди "
                                                 "допустимых вариантов.")
                                            ]
                                        }
                                    }
                                }
                            },
                            'status_code': {
                                'type': 'integer',
                                'example': 400
                            }
                        }
                    }
                }
            }
        }

        return responses


class GUIAutoSchema(AutoSchema):
    """Авто схема с поддержкой uds.core.ui.gui элементов

    Позволяет добавлять в спецификацию API метода описание
    полей классов-наследников UserInterface.

    Пример: termidesk.api.webui.views.osmanagers.OSManagerAutoSchema
    """

    def get_gui_types(self) -> list:
        """Возвращает список классов-наследников UserInterface"""
        raise NotImplementedError

    def get_components(self, path, method):
        """Добавляет компоненты на основе списка классов-наследников UserInterface"""
        components = super().get_components(path, method)

        for gui_type in self.get_gui_types():
            component = self.get_gui_component(gui_type, components)
            if component:
                components[gui_type.type()] = component

        return components

    def get_gui_component(self, gui_type, components) -> dict:
        """Возвращает компонент на основе класса-наследника UserInterface

        Пример:
            gui_fields = self.map_gui_fields(
                gui_type.guiDescription(),
                deepcopy(components['OSManager']),
            )
            return gui_fields
        """
        raise NotImplementedError

    def map_gui_fields(self, fields, content=None):
        """Преобразовывает словарь из uds.core.ui.gui
        элементов в словарь-схему для сваггера
        """
        content = content or {}
        required = content.setdefault('required', [])
        properties = content.setdefault('properties', {})

        for field in fields:
            if field['gui']['type'] == gui.InputField.HIDDEN_TYPE:
                continue
            if field['gui']['required']:
                required.append(field['name'])

            schema = self.map_gui_field(field)
            properties[field['name']] = schema

        return content

    def map_gui_field(self, field):
        """Преобразовывает uds.core.ui.gui элемент
        в схему для сваггера
        """
        schema = {}
        text_types = (
            gui.InputField.TEXT_TYPE,
            gui.InputField.TEXTBOX_TYPE,
            gui.InputField.PASSWORD_TYPE,
            gui.InputField.INFO_TYPE,
        )
        choice_types = (
            gui.InputField.CHOICE_TYPE,
            gui.InputField.EDITABLE_LIST,
            gui.InputField.IMAGECHOICE_TYPE,
        )

        if field['gui']['type'] == gui.InputField.HIDDEN_TYPE:
            raise TypeError('Hidden Type must be skipped')

        if field['gui']['type'] == gui.InputField.INFO_TYPE:
            field['gui']['rdonly'] = True

        if field['gui']['type'] in text_types:
            schema = {'type': 'string'}

        elif field['gui']['type'] == gui.InputField.NUMERIC_TYPE:
            schema = {
                'type': 'integer',
                'maximum': field['gui']['maxValue'],
                'minimum': field['gui']['minValue'],
            }

        elif field['gui']['type'] in choice_types:
            schema = {
                'type': 'string',
                'enum': [value['id'] for value in field['gui']['values']],
            }

        elif field['gui']['type'] == gui.InputField.MULTI_CHOICE_TYPE:
            schema = {
                'type': 'array',
                'items': {
                    'type': 'string',
                    'enum': [value['id'] for value in field['gui']['values']],
                },
            }

        elif field['gui']['type'] == gui.InputField.CHECKBOX_TYPE:
            schema = {
                'type': 'string',
                'enum': [gui.TRUE, gui.FALSE],
            }

        elif field['gui']['type'] == gui.InputField.DATE_TYPE:
            schema = {
                'type': 'string',
                'format': 'date',
            }

        self._update_additional_fields(field, schema)

        return schema

    @staticmethod
    def _update_additional_fields(field, schema):
        """Добавляет в схему дополнительные общие атрибуты из gui поля"""

        if field['gui']['length']:
            schema['maxLength'] = field['gui']['length']

        if field['gui']['rdonly']:
            schema['readOnly'] = field['gui']['rdonly']

        if field['gui']['defvalue']:
            schema['default'] = field['gui']['defvalue']

        if field['gui']['tooltip']:
            schema['description'] = field['gui']['tooltip']
        elif field['gui']['label']:
            schema['description'] = field['gui']['label']
