import logging

from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.exceptions import APIException
from rest_framework.fields import empty

from uds.core import osmanagers
from uds.core.util import permissions
from uds.models import Tag
from uds.models.OSManager import OSManager

logger = logging.getLogger(__name__)

ON_LOGOUT_VALUES = ['keep', 'remove']


class TagsSlugRelatedField(serializers.SlugRelatedField):
    """
    Поле для тэгов менеджеров ОС

    Так как тэги сейчас не используются, то везде передается значение [''].
    Для SlugRelatedField тэг с таким значением должен присутствовать в БД,
    иначе будет ошибка. Для избежания этой ошибки, нужна доработка ниже.
    """

    def to_internal_value(self, data):
        if data == '':
            instance, _created = Tag.objects.get_or_create(tag=data)
            return instance
        return super().to_internal_value(data)


class OSManagerOverviewSerializer(serializers.ModelSerializer):
    """ Сериализатор списка доступных параметров гостевых ОС /overview. """
    id = serializers.CharField(source='uuid', required=False, read_only=True)
    tags = TagsSlugRelatedField(many=True, slug_field='tag', queryset=Tag.objects.all())
    deployed_count = serializers.IntegerField(source='deployedServices.count', read_only=True)
    servicesTypes = serializers.ListField(source='getType.servicesType', read_only=True)
    type = serializers.CharField(source='getType.type', read_only=True)
    comments = serializers.CharField(
        allow_blank=True,
        label=_('Комментарий'),
        help_text=_('Комментарий для объекта'),
    )
    name = serializers.CharField(
        label=_('Название'),
        help_text=_('Имя объекта'),
    )
    permission = serializers.SerializerMethodField()

    class Meta:
        model = OSManager
        fields = ['id', 'name', 'tags', 'deployed_count', 'type', 'servicesTypes', 'comments', 'permission']

    def get_permission(self, instance):
        request = self.context.get('request')
        return permissions.getEffectivePermission(request.user, instance)


class OSManagerSerializer(OSManagerOverviewSerializer):
    """ Расширенный сериализатор для работы с детальной параметов гостевых ОС. """
    onLogout = serializers.ChoiceField(source="getInstance.valuesDict.onLogout", choices=ON_LOGOUT_VALUES)
    maxSessionTime = serializers.IntegerField(
        source="getInstance.valuesDict.maxSessionTime",
        required=False,
    )

    def validate_maxSessionTime(self, value):
        """Валидация длительности сессии.

        Корректными значениями параметра `Длительность сессии пользователя`
        являются -1 и от 60 до 604800 (3600*24*7)
        """

        if not osmanagers.OSManager.is_session_duration_valid(value):
            raise serializers.ValidationError(
                '"{}" is not acceptable value for maxSessionTime. '
                'Valid are -1, 60-604800'.format(value)
            )

        return value

    class Meta:
        model = OSManager
        fields = ['id', 'name', 'tags', 'deployed_count', 'type', 'servicesTypes', 'comments', 'permission',
                  'onLogout', 'data_type', 'maxSessionTime', ]
        extra_kwargs = {
            'data_type': {'write_only': True, 'required': False},
        }

    def __init__(self, instance=None, data=empty, **kwargs):
        super().__init__(instance=instance, data=data, **kwargs)

        # проверка на попытку обновления
        # если обновление - то нужно заполнить данными из instance
        if instance is not None and data is not empty:
            self.initial_data = self.get_full_data(self.initial_data)

    def get_full_data(self, data: dict):
        """ Получить список всех значений полей. """

        for name, value in self.instance.getInstance().valuesDict().items():
            if data.get(name) is None:
                data[name] = value
        return data

    def validate(self, attrs):
        logger.debug('Before {0}'.format(attrs))

        if not self._has_tags_param():
            raise serializers.ValidationError('Not enough «tags» parameter')

        data_type = attrs.get('data_type', self.instance.data_type if self.instance else '')
        type = osmanagers.factory().lookup(data_type)
        if type is None:
            raise serializers.ValidationError(
                {'data_type': 'This is not exists «{}» type'.format(data_type)})

        try:
            instance_type_params = attrs['getInstance']['valuesDict']
        except KeyError:
            raise serializers.ValidationError('Not enough parameters to save')

        on_logout = instance_type_params.get('onLogout')
        if self.instance is None and on_logout not in ON_LOGOUT_VALUES:
            raise serializers.ValidationError(
                {'onLogout': '«{}» is not acceptable value for onLogout'.format(on_logout)})

        return super().validate(attrs)

    def create(self, validated_data):
        name = validated_data.get('name')
        comments = validated_data.get('comments')
        data_type = validated_data.get('data_type')

        osmanager = OSManager(name=name, comments=comments, data_type=data_type)
        return self.validate_object_and_save(osmanager)

    def update(self, instance: OSManager, validated_data: dict):
        instance.name = validated_data.get('name', instance.name)
        instance.comments = validated_data.get('comments', instance.comments)
        instance.data_type = validated_data.get('data_type', instance.data_type)

        return self.validate_object_and_save(instance)

    def validate_object_and_save(self, osmanager: OSManager):
        try:
            self._validate_session_duration(osmanager)

            osmanager.data = osmanager.getInstance(self.initial_data).serialize()
            osmanager.save()
        except KeyError as key_error:
            raise APIException(
                'The field {} required for «{}» type'.format(key_error, osmanager.data_type))
        except osmanagers.OSManager.ValidationException as validate_error:
            raise APIException(validate_error)

        # Now if tags, update them
        tags = self.initial_data.get('tags')
        if tags is not None:
            logger.debug('Updating tags: {}'.format(tags))
            osmanager.tags.set([Tag.objects.get_or_create(tag=val)[0] for val in tags])

        return osmanager

    def _has_tags_param(self):
        return 'tags' in self.initial_data.keys()

    def _validate_session_duration(self, _osmanager):
        """Валидация значения параметра `Длительность сессии пользователя`"""
        session_duration = self.initial_data.get('maxSessionTime')
        # Если параметр не передан с запросом, например, из-за отключения параметра в настройках,
        # то пытаемся получить сохраненное значение
        if session_duration is None:
            session_duration = _osmanager.getInstance().maxSession()

        session_duration = int(session_duration)

        self.validate_maxSessionTime(session_duration)

        self.initial_data['maxSessionTime'] = session_duration
