# -*- encoding: utf-8 -*-
import logging
from django.utils.translation import gettext as _
from rest_framework import serializers
from rest_framework.fields import empty
from rest_framework.validators import UniqueValidator, ValidationError
from termidesk.core.util.utils import hide_sensitive_data
from uds.models.Provider import Provider
from uds.models import UserService
from uds.core.util.State import State
from uds.core.services import factory
from uds.core.util import permissions

logger = logging.getLogger(__name__)

OVIRT_VERSION = ['4']
PLATFORM_VERSION = ['2', '6']


class ProviderOverviewSerializer(serializers.ModelSerializer):
    """
    Сериализатор списка доступных параметров поставщиков ресурсов /overview.
    """
    id = serializers.CharField(source='uuid', required=False)
    name = serializers.CharField(
        required=True,
        validators=[
            UniqueValidator(
                queryset=Provider.objects.all(),
                message=_('Element already exists (duplicate key error)')
            )
        ],
        label='Название',
        help_text='Название поставщика ресурсов',
    )
    comments = serializers.CharField(
        required=False,
        label='Комментарий',
        allow_blank=True,
        help_text='Комментарий поставщика ресурсов',
    )
    type = serializers.CharField(source='getType.type', read_only=True)

    permission = serializers.SerializerMethodField()
    tags = serializers.SerializerMethodField()
    type_name = serializers.SerializerMethodField()
    services_count = serializers.SerializerMethodField()
    maintenance_mode = serializers.SerializerMethodField()
    offers = serializers.SerializerMethodField()
    user_services_count = serializers.SerializerMethodField()

    class Meta:
        model = Provider
        fields = ['id', 'name', 'comments', 'type', 'permission', 'tags', 'type_name',
                  'services_count', 'maintenance_mode', 'offers', 'user_services_count']

    def __init__(self, instance=None, data=empty, **kwargs):
        super().__init__(instance=instance, data=data, **kwargs)

        # проверка на попытку обновления
        # если обновление - то нужно заполнить данными из instance
        if instance is not None and data is not empty:
            self.initial_data = self.get_full_data(self.initial_data)

    def get_full_data(self, data: dict):
        """ Получить список всех значений полей. """
        for name, value in self.instance.getInstance().valuesDict().items():
            if data.get(name) is None:
                data[name] = value
        return data

    def get_user_services_count(self, instance):
        """Получение информации о рабочих местах поставщиков"""
        user_services_count = UserService.objects.filter(
            deployed_service__service__provider=instance
            ).exclude(state__in=(State.REMOVED, State.ERROR)).count()
        return user_services_count

    def get_services_count(self, instance):
        """Получение информации о  количестве шаблонов поставщиков"""
        if not isinstance(instance, Provider):
            return []
        return instance.services.count()

    def get_maintenance_mode(self, instance):
        """Получение информации о техобслуживании"""
        if not isinstance(instance, Provider):
            return []
        return instance.maintenance_mode

    def get_offers(self, instance):
        if not isinstance(instance, Provider):
            return []
        offers = [{
            'name': _(t.name()),
            'type': t.type(),
            'description': _(t.description()),
            'icon': t.icon().replace('\n', '')} for t in instance.getType().getServicesTypes()]
        return offers

    def get_permission(self, instance):
        """Получение разрешений для пользователя из запроса"""
        request = self.context.get('request')
        return permissions.getEffectivePermission(request.user, instance)

    def get_tags(self, instance):
        """Получение меток (тэгов) для поставщиков ресурсов"""
        if not isinstance(instance, Provider):
            return ''
        return [tag.tag for tag in instance.tags.all()]

    def get_type(self, instance):
        """Получение типа поставщиков ресурсов"""
        if not isinstance(instance, Provider):
            return ''
        return self.validated_data.get("data_type")

    def get_type_name(self, instance):
        """Получение имени типа поставщиков ресурсов"""
        if not isinstance(instance, Provider):
            return ''
        return instance.getType().name()

    def maintenance(self, instance):
        """Изменение параметра 'техобслуживание' """

        instance.maintenance_mode = not instance.maintenance_mode
        instance.save()

        return instance


class ProviderSerializer(serializers.ModelSerializer):
    """ Расширенный сериализатор для работы с детальной параметов поставщиков ресурсов. """

    id = serializers.CharField(source='uuid', required=False)
    name = serializers.CharField(
        required=True,
        validators=[
            UniqueValidator(
                queryset=Provider.objects.all(),
                message=_('Element already exists (duplicate key error)')
            )
        ],
        label='Название',
        help_text='Название поставщика ресурсов',
    )
    type = serializers.CharField(source='getType.type', read_only=True, required=False)
    comments = serializers.CharField(
        required=False,
        label='Комментарий',
        allow_blank=True,
        help_text='Комментарий поставщика ресурсов',
    )
    maintenance_mode = serializers.SerializerMethodField()
    permission = serializers.SerializerMethodField()
    tags = serializers.SerializerMethodField()

    class Meta:
        model = Provider
        fields = ['id', 'maintenance_mode', 'tags', 'name', 'permission', 'comments', 'type']

    def __init__(self, instance=None, data=empty, **kwargs):
        super().__init__(instance=instance, data=data, **kwargs)

        # проверка на попытку обновления
        # если обновление - то нужно заполнить данными из instance
        if instance is not None and data is not empty:
            self.initial_data = self.get_full_data(self.initial_data)

    def get_full_data(self, data: dict):
        """ Получить список всех значений полей. """
        for name, value in self.instance.getInstance().valuesDict().items():
            if data.get(name) is None:
                data[name] = value
        return data
    def get_maintenance_mode(self, instance):
        """Получение информации о техобслуживании"""
        if not isinstance(instance, Provider):
            return []
        return instance.maintenance_mode

    def get_offers(self, instance):
        if not isinstance(instance, Provider):
            return []
        offers = [{
            'name': _(t.name()),
            'type': t.type(),
            'description': _(t.description()),
            'icon': t.icon().replace('\n', '')} for t in instance.getType().getServicesTypes()]
        return offers

    def get_permission(self, instance):
        """Получение разрешений для пользователя из запроса"""
        request = self.context.get('request')
        return permissions.getEffectivePermission(request.user, instance)

    def get_tags(self, instance):
        """Получение меток (тэгов) для поставщиков ресурсов"""
        if not isinstance(instance, Provider):
            return ''
        return [tag.tag for tag in instance.tags.all()]

    def create(self, validated_data):
        prov_name = validated_data.get('name')
        prov_data_type = validated_data.get('data_type')
        prov_comments = validated_data.get('comments')

        provider = Provider(
            name=prov_name,
            comments=prov_comments,
            data_type=prov_data_type,
        )
        try:
            prov_type = factory().lookup(prov_data_type)
            attributes = {}
            for key in prov_type._gui.keys():
                attributes.update({key: validated_data.get(key)})
            provider.data = provider.getInstance(attributes).serialize()
        except Exception as error:
            print("Error", error)

        provider.save()
        return provider

    def update(self, instance, validated_data):
        instance.name = validated_data['name']
        instance.comments = validated_data['comments']
        prov_type = factory().lookup(validated_data['data_type'])
        attrs = {}
        for key in prov_type._gui.keys():
            attrs.update({key: validated_data[key]})

        instance.data = instance.getInstance(attrs).serialize()
        instance.save()

        return instance

    def validate(self, attrs):
        data = self.context.get('request').data
        data_type = data['data_type']

        type_ = factory().lookup(data_type)
        if type_ is None:
            raise serializers.ValidationError(
                {'data_type': 'This is not exists «{}» type'.format(data_type)})

        # инициализации сущности, при недостатке данных выпадет ошибка
        try:
            prov_type = factory().lookup(data_type)
            prov_obj = prov_type(self.initial_data)
        except KeyError as key:
            raise serializers.ValidationError(
                {'The field {} required for «{}» type'.format(key, data_type)}
            )

        custom_fields_list = self.get_custom_fields(prov_type)
        for custom_field_name in custom_fields_list:
            try:
                attrs[custom_field_name] = self.validate_custom_field(custom_field_name)
            except ValidationError as validation_exception:
                raise validation_exception
            except KeyError:
                continue

        # Дополнительная проверка обязательных полей нужна для случая перехода
        # с обращения по ключу на обращение через метод get().
        # При установке свойств объекта заданного типа, мы все-равно должны быть увенрены,
        # что все обязательные поля заданы.
        for key in prov_obj.get_required_fields():
            if key not in data:
                raise serializers.ValidationError('Key {} not found'.format(key))

        logger.debug('Before {0}'.format(hide_sensitive_data(attrs)))

        try:
            attrs = data
        except Exception as error:
            raise serializers.ValidationError('Invalid attribute: ' + str(error))
        return super().validate(attrs)

    def get_custom_fields(self, prov_type=None):
        """
        Получение списка полей, уникальных для данного типа поставщика
        :param prov_type: строка с типом поставщика
        :return: список имен уникальных полей для данного типа поставщика.
        """
        custom_fields_list = []
        if prov_type:
            prov_instance = prov_type(self.initial_data)
            custom_fields_list = [x['name'] for x in prov_instance.guiDescription()]
        return custom_fields_list

    def validate_custom_field(self, field_name):
        """Проверка полей, специфичных для поставщика"""

        provider_gui_description = factory().lookup(self.initial_data['data_type']).guiDescription()
        target_field = next((x for x in provider_gui_description if x["name"] == field_name), {})
        is_required_field = target_field.get('gui', {}).get('required', False)
        target_field_value = self.initial_data.get(field_name)

        if is_required_field and target_field_value == '':
            raise ValidationError(
                "The field {required_field} is empty.".format(required_field=field_name))

        elif is_required_field and target_field_value is None:
            raise ValidationError("The field {required_field} not found."
                                  .format(required_field=field_name))

        else:
            return self.initial_data[field_name]
