import logging

from rest_framework import serializers
from rest_framework.exceptions import APIException
from rest_framework.fields import empty
from uds.core.util import permissions
from uds.core.util.State import State
from uds.core.services import Service as coreService
from uds.models.Service import Service
from uds.models.UserService import UserService

logger = logging.getLogger(__name__)


class ServiceInfoSerializer(serializers.Serializer):
    icon = serializers.SerializerMethodField()
    needs_publication = serializers.BooleanField(source='publicationType')
    max_deployed = serializers.IntegerField(source='maxDeployed')
    uses_cache = serializers.BooleanField(source='usesCache')
    uses_cache_l2 = serializers.BooleanField(source='usesCache_L2')
    cache_tooltip = serializers.CharField(source='cacheTooltip')
    cache_tooltip_l2 = serializers.CharField(source='cacheTooltip_L2')
    needs_manager = serializers.BooleanField(source='needsManager')
    allowedProtocols = serializers.ListField()
    servicesTypeProvided = serializers.ListField()
    must_assign_manually = serializers.BooleanField(source='mustAssignManually')

    def get_icon(self, instance):
        icon = instance.icon()
        if not isinstance(icon, str):
            icon = icon.decode('utf-8')

        return icon.replace('\n', '')


class ServiceListSerializer(serializers.ModelSerializer):
    id = serializers.CharField(source='uuid', required=False, read_only=True)
    comments = serializers.CharField(allow_blank=True)
    type = serializers.CharField(source='data_type', read_only=True)
    type_name = serializers.CharField(source='getType.name', read_only=True)
    deployed_services_count = serializers.IntegerField(source='deployedServices.count', read_only=True)
    user_services_count = serializers.SerializerMethodField()
    maintenance_mode = serializers.BooleanField(source='provider.maintenance_mode', read_only=True)
    permission = serializers.SerializerMethodField()
    tags = serializers.SerializerMethodField()

    class Meta:
        model = Service
        fields = ['id', 'name', 'comments', 'type', 'type_name', 'deployed_services_count',
                  'user_services_count', 'maintenance_mode', 'permission', 'tags']

    def get_tags(self, instance):
        if not isinstance(instance, Service):
            return []
        return [tag.tag for tag in instance.tags.all()]

    def get_permission(self, instance):
        if not isinstance(instance, Service):
            return []
        request = self.context.get('request')
        return permissions.getEffectivePermission(request.user, instance)

    def get_user_services_count(self, instance):
        if not isinstance(instance, Service):
            return 0
        return (UserService.objects
                .filter(deployed_service__service=instance)
                .exclude(state__in=(State.REMOVED, State.ERROR))
                .count())


class ServiceDetailSerializer(ServiceListSerializer):
    info = ServiceInfoSerializer(source='getType', read_only=True)
    data_type = serializers.CharField(write_only=True)

    class Meta:
        model = Service
        fields = ['id', 'name', 'data_type', 'comments', 'type', 'type_name', 'deployed_services_count',
                  'user_services_count', 'maintenance_mode', 'permission', 'info', 'tags']

    def __init__(self, instance=None, data=empty, **kwargs):
        self.provider = kwargs.pop('provider')
        super().__init__(instance=instance, data=data, **kwargs)

        # проверка на попытку обновления
        # если обновление - то нужно заполнить данными из instance
        if instance is not None and data is not empty:
            self.initial_data = self.get_full_data(self.initial_data)

    def get_full_data(self, data: dict):
        """ Получить список всех значений полей. """

        for name, value in self.instance.getInstance().valuesDict().items():
            if data.get(name) is None:
                data[name] = value
        return data

    def validate(self, attrs):
        logger.debug('Before {0}'.format(attrs))

        data_type = attrs.get('data_type', self.instance.data_type if self.instance else '')
        type_instance = self.provider.getType().getServiceByType(data_type)
        if type_instance is None:
            raise serializers.ValidationError(
                {'data_type': 'This is not exists «{}» type'.format(data_type)})

        self.__check_required_fields(type_instance)

        return super().validate(attrs)

    def save(self, **kwargs):
        error = None
        service = None

        try:
            if self.instance is None:
                data = self.validated_data

                if data['data_type'] == 'pksvbrestFullLiveService':
                    # Запрет создания шаблона ПК СВ БРЕСТ типа pksvbrestFullLiveService,
                    # так как он устарел. Этот запрет будет удален после удаления типа
                    # pksvbrestFullLiveService из плагинов в задаче [TSRV-752]
                    raise serializers.ValidationError({
                        'data_type': 'Deprecated template type. Can\'t create new objects'
                    })

                data['provider_id'] = self.provider.id
                service = self.provider.services.create(**data)
            else:
                service = self.instance
                service.__dict__.update(self.validated_data)

            service.data = service.getInstance(self.initial_data).serialize()
            service.save()

        except APIException:
            # Не мешаем DRF корректно обрабатывать его собственные ошибки
            raise

        except coreService.ValidationException as err:
            error = 'Ошибка ввода: {0}'.format(err)
        except Exception as err:
            error = err

        if error is not None:
            if self.instance is None and service is not None:
                service.delete()
            logger.exception('Saving Service')
            raise APIException(error)

        logger.debug('Saving group {0} / {1}'.format(self.provider, self.validated_data.get('name')))
        self.instance = service
        return self.instance

    def __check_required_fields(self, type_instance):
        """ Проверка всех обязательных полей для переданного типа. """

        required_fields = (key for key, params in type_instance._gui.items() if
                           params._data.get('required') is True)
        for key in required_fields:
            if key not in self.initial_data:
                raise serializers.ValidationError(
                    {key: 'The field {} required for «{}» type'.format(key, type_instance)}
                )
