# -*- encoding: utf-8 -*-

from django.contrib.auth.models import Group
from guardian.models.models import GroupObjectPermission
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from termidesk.api.decorators import has_role_permission
from termidesk.api.webui.serializers.acl.entity import EntitySerializer
from termidesk.api.webui.serializers.acl.workplace import WorkplaceEntitySerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.api.webui.serializers.type import TypeSerializer
from termidesk.models.UserPermission import USER_ROLES_MANAGE
from uds.core.ui.UserInterface import gui


class AccessControlList(CommonViewSet):
    """
    gui:
        Спецификация полей GUI для пользовательских разрешений к конкретным объектам.
    retrieve:
        Получить информацию об объектном разрешении
    get_types:
        Получить типы объектов, доступных для ассоциации
    overview:
        Список всех доступных объектных разрешений
    create:
        Создание разрешения к конкретному объекту в БД
    update:
        Обновить параметры объектного разрешения
    delete:
        Удалить объектное разрешение
    """
    lookup_field = 'id'
    model = GroupObjectPermission
    serializer_class = EntitySerializer

    ENTITIES = {
        'workplace': WorkplaceEntitySerializer,
    }

    def get_queryset(self):
        return self.model.objects.select_related('group', 'permission', 'content_type').all()

    @has_role_permission(USER_ROLES_MANAGE)
    def retrieve(self, request, *args, **kwargs):
        return super().retrieve(request, *args, **kwargs)

    @has_role_permission(USER_ROLES_MANAGE)
    def get_types(self, request, *args, **kwargs):
        data = []
        for entity, obj in self.ENTITIES.items():
            data.append({'name': obj.title,
                         'type': entity,
                         'description': obj.description,
                         'icon': ''})

        serializer = TypeSerializer(data=data, many=True)

        return Response(sorted(serializer.initial_data, key=lambda f: f['name']))

    def get_gui_params(self, **kwargs):
        serializer_class = self.get_serializer_class()
        objects = serializer_class.get_gui_objects()

        groups = (gui.choiceItem(group.id, group.name) for group in Group.objects.order_by('name').all())
        gui_represent = self.add_field(self.add_default_fields([], []),
                                       {'name': 'group_id',
                                        'label': 'Роль',
                                        'values': groups,
                                        'tooltip': '',
                                        'type': gui.ChoiceField.CHOICE_TYPE,
                                        'order': 101,
                                        'required': True, })

        permissions = serializer_class.get_gui_allow_permissions()
        self.add_field(gui_represent, {'name': 'permission_id',
                                       'label': 'Пользовательское разрешение',
                                       'values': permissions,
                                       'tooltip': '',
                                       'type': gui.ChoiceField.CHOICE_TYPE,
                                       'order': 102,
                                       'required': True, })

        self.add_field(gui_represent,
                       {'name': 'object_uuid',
                        'label': 'Объект',
                        'values': objects,
                        'tooltip': '',
                        'type': gui.ChoiceField.CHOICE_TYPE,
                        'order': 103,
                        'required': True, },
                       )

        return gui_represent

    @has_role_permission(USER_ROLES_MANAGE)
    def create(self, request, *args, **kwargs):
        return super().create(request, *args, **kwargs)

    @has_role_permission(USER_ROLES_MANAGE)
    def update(self, request, *args, **kwargs):
        return super().update(request, *args, **kwargs)

    @has_role_permission(USER_ROLES_MANAGE)
    def overview(self, request, *args, **kwargs):
        response = super().list(request, *args, **kwargs)

        # добавление удобочитаемого типа объекта
        types = self.__get_list_types_object_with_title()
        for permission in response.data:
            permission['type_title'] = types[permission['type_object']]['title']
            permission['type_object'] = types[permission['type_object']]['type']

        return response

    @has_role_permission(USER_ROLES_MANAGE)
    def delete(self, request, *args, **kwargs):
        return super().destroy(request, *args, **kwargs)

    def __get_list_types_object_with_title(self):
        """
        Формирование списка объектов следующего вида:
        {
            'название класса': {
                'title': 'Подпись для отображения в списке',
                'type': 'Тип для получения gui',
            },
        }
        """
        entities = {}
        for type_, serialize_class in self.ENTITIES.items():
            entities[serialize_class.model.__name__] = {
                'title': serialize_class.title, 'type': type_
            }
        return entities

    def get_serializer_class(self):
        if self.request:
            if self.request.data:
                entity = self.request.data.get('data_type')
            else:
                # TODO: url schema should not use parameter as query root (/api/webui/acl/gui/<workplace>)
                # it should be query param (/api/webui/acl/gui?data_type=workplace) for correct processing
                entity = self.request.parser_context['kwargs'].get('data_type')
        else:
            # request not initialized only for openapi doc generation
            entity = None

        if not entity:
            # enclosed data type not specified in request
            return self.serializer_class

        if entity not in self.ENTITIES:
            raise ValidationError(
                'Переданный тип объекта "{}" не доступен для создания разрешения'.format(entity)
            )
        return self.ENTITIES.get(entity)
