# -*- encoding: utf-8 -*-

import logging
from abc import ABCMeta, abstractmethod

import six
import warnings
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import PermissionDenied
from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.viewsets import ModelViewSet
from termidesk.api.permissions import AccessPermission
from termidesk.core.util.utils import hide_sensitive_data
from termidesk.rest.handlers import (
    CREATE_ACTION, DELETE_ACTION, READ_ACTION, UPDATE_ACTION, set_crud_event, set_crud_event_from_viewset
)
from uds.core.ui import gui as GUI

logger = logging.getLogger(__name__)

OK = 'ok'


class CommonViewSet(ModelViewSet):
    """ Класс, который хранит методы по частичной генерации документации
     и другие общие свойства и методы."""
    __metaclass__ = ABCMeta
    permission_classes = [AccessPermission]
    schema = AutoSchema()
    serializer_classes = {}

    # Метод позволяет частично сгенерировать документацию на базе сериалайзера
    def get_serializer_context(self):
        return {
            'request': self.request,
            'format': self.format_kwarg,
            'view': self
        }

    def get_serializer_class(self, *args, **kwargs):
        return self.serializer_classes.get(self.action, self.serializer_class)

    def get_queryset(self):
        return self.model.objects.all()

    def get_object(self):
        # Perform the lookup filtering.
        if self.lookup_field == 'uuid':
            if self.lookup_field in self.kwargs:
                self.kwargs[self.lookup_field] = self.kwargs[self.lookup_field].lower()
        return super().get_object()

    @staticmethod
    def add_field(gui, field):
        """
        Add a field to a "gui" description.
        This method checks that every required field element is in there.
        If not, defaults are assigned
        :param gui: List of "gui" items where the field will be added
        :param field: Field to be added (dictionary)
        """
        warnings.warn(
            ('Method add_field is deprecated.'
             'Use termidesk.api.webui.helpers.GuiHelper.add_field instead.'),
            DeprecationWarning, stacklevel=2
        )

        v = {
            'name': field.get('name', ''),
            'value': '',
            'gui': {
                'required': field.get('required', False),
                'defvalue': field.get('value', ''),
                'value': field.get('value', ''),
                'minValue': field.get('minValue', '987654321'),
                'label': field.get('label', ''),
                'length': field.get('length', 128),
                'multiline': field.get('multiline', 0),
                'tooltip': field.get('tooltip', ''),
                'rdonly': field.get('rdonly', False),
                'type': field.get('type', GUI.InputField.TEXT_TYPE),
                'order': field.get('order', 0),
                'values': field.get('values', [])
            }
        }
        if 'tab' in field:
            v['gui']['tab'] = field['tab']
        gui.append(v)
        return gui

    def add_default_fields(self, gui, fields):
        """
        Adds default fields (based in a list) to a "gui" description
        :param gui: Gui list where the "default" fielsds will be added
        :param fields: List of fields names requested to be added.
        Valid values are 'name', 'comments', 'priority' and 'small_name'
        """
        warnings.warn(
            ('Method add_default_fields is deprecated.'
             'Use termidesk.api.webui.helpers.GuiHelper.add_default_fields instead.'),
            DeprecationWarning, stacklevel=2
        )

        if 'tags' in fields:
            self.add_field(gui, {
                'name': 'tags',
                'label': _('Теги'),
                'type': 'taglist',
                'tooltip': _('Теги для данного объекта'),
                'order': 0 - 101,
            })

        if 'name' in fields:
            self.add_field(gui, {
                'name': 'name',
                'required': True,
                'label': _('Название'),
                'length': 128,
                'tooltip': _('Имя объекта'),
                'order': 0 - 100,
            })

        if 'comments' in fields:
            self.add_field(gui, {
                'name': 'comments',
                'label': _('Комментарий'),
                'tooltip': _('Комментарий для объекта'),
                'length': 256,
                'order': 0 - 99,
            })

        if 'priority' in fields:
            self.add_field(gui, {
                'name': 'priority',
                'type': 'numeric',
                'label': _('Приоритет'),
                'tooltip': _('Приоритет для данного объекта (чем меньше число, тем выше приоритет)'),
                'required': True,
                'value': 1,
                'length': 4,
                'order': 0 - 97,
            })

        if 'small_name' in fields:
            self.add_field(gui, {
                'name': 'small_name',
                'type': 'text',
                'label': _('Метка'),
                'tooltip': _('Метка (короткое имя) для данного объекта'),
                'required': True,
                'length': 128,
                'order': 0 - 96,
            })

        return gui

    def type_info(self, type_):  # pylint: disable=no-self-use
        '''
        Returns info about the type
        In fact, right now, it returns an empty dict, that will be extended by typeAsDict
        '''
        return {}

    def type_as_dict(self, type_):
        '''
        Returns a dictionary describing the type (the name, the icon, description, etc...)
        '''
        res = self.type_info(type_)
        icon = type_.icon()
        icon = icon.replace('\n', '')
        res.update({
            'name': _(type_.name()),
            'type': type_.type(),
            'description': _(type_.description()),
            'icon': icon
        })
        if hasattr(type_, 'group'):
            res['group'] = _(type_.group)  # Add group info is it is contained
        return res

    def access_denied(self, message=None):
        raise PermissionDenied(message or _('Access denied'))

    def success(self):
        '''
        Utility method to be invoked for simple methods that returns nothing in fact
        '''
        request_data = hide_sensitive_data(self.request.data)
        logger.debug('Returning success on {} {}'.format(self.__class__, request_data))
        return OK

    def fill_intance_fields(self, item, res):  # pylint: disable=no-self-use
        '''
        For Managed Objects (db element that contains a serialized object),
        fills a dictionary with the "field" parameters values.
        For non managed objects, it does nothing
        :param item: Item to extract fields
        :param res: Dictionary to "extend" with instance key-values pairs
        '''
        if hasattr(item, 'getInstance'):
            i = item.getInstance()
            i.initGui()  # Defaults & stuff
            # 2to3: iteritems() changed to items()
            for key, value in i.valuesDict().items():
                if type(value) in (six.text_type, six.binary_type):
                    # Translate "true" & "false" to True & False (booleans)
                    value = {"true": True, "false": False}.get(value, value)
                logger.debug('{0} = {1}'.format(key, value))
                res[key] = value
        return res

    def gui(self, request, *args, **kwargs):
        gui = self.get_gui_params(**kwargs)
        return Response(sorted(gui, key=lambda f: f['gui']['order']))

    @abstractmethod
    def get_gui_params(self, **kwargs):
        """ Возвращает основные параметры для gui. Метод обязательный к перекрытию. """


class ChildCommonViewSet(CommonViewSet):
    """ Класс содержит общие методы для работы с подсущностями. """
    search_str = 'search'

    def _set_read_instance_audit(self, instance):
        """ Фиксация события чтения сущности. """
        subtype = instance.data_type if hasattr(instance, 'data_type') else ''

        set_crud_event(
            self.request.user.manager.name,
            self.request.user.name,
            self.request.ip,
            instance.__class__.__name__,
            instance.uuid,
            subtype,
            instance.name,
            READ_ACTION
        )

    def _set_update_parents_audit(self, parent):
        """ Фиксация события обновления родительской сущности. """
        subtype = parent.data_type if hasattr(parent, 'data_type') else ''

        set_crud_event(
            self.request.user.manager.name,
            self.request.user.name,
            self.request.ip,
            parent.__class__.__name__,
            parent.uuid,
            subtype,
            parent.name,
            UPDATE_ACTION
        )


class CrudCommonViewSet(CommonViewSet):
    """Класс с созданием эвентов для базовых методов REST API"""

    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        set_crud_event_from_viewset(self, serializer, READ_ACTION)
        return Response(serializer.data)

    def full_retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        set_crud_event_from_viewset(self, serializer, READ_ACTION)
        return Response(serializer.get_full_data(serializer.data))

    def perform_create(self, serializer):
        super().perform_create(serializer)
        set_crud_event_from_viewset(self, serializer, CREATE_ACTION)

    def perform_update(self, serializer):
        super().perform_update(serializer)
        set_crud_event_from_viewset(self, serializer, UPDATE_ACTION)

    def perform_destroy(self, instance):
        set_crud_event(
            self.request.user.manager.name,
            self.request.user.name,
            self.request.ip,
            self.model.__name__,
            instance.uuid,
            getattr(instance, 'type', getattr(instance, 'data_type', '')),
            instance.name,
            DELETE_ACTION
        )
        super().perform_destroy(instance)
