import logging
from copy import deepcopy

from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet
from rest_framework.mixins import ListModelMixin
from termidesk.api.permissions import AccessPermission

from termidesk.api.decorators import has_role_permission
from termidesk.api.webui.schemas import GUIAutoSchema
from termidesk.api.webui.serializers.osmanager import OSManagerOverviewSerializer, OSManagerSerializer
from termidesk.api.webui.serializers.type import TypeSerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet, CrudCommonViewSet
from termidesk.models.UserPermission import OSMANAGER_VIEW, OSMANAGER_CHANGE, OSMANAGER_DELETE, \
    OSMANAGER_CREATE

from uds.core.osmanagers import factory
from uds.models.OSManager import OSManager

logger = logging.getLogger(__name__)


class OSManagersGUI(CommonViewSet):
    """
    API VIEW для элементов GUI

    gui:
        Спецификация полей GUI определенного типа для параметров гостевой ОС.
    """
    model = OSManager
    serializer_class = OSManagerSerializer

    def get_gui_params(self, **kwargs):
        type_osmanager = kwargs.get('type')
        try:
            fields = factory().lookup(type_osmanager).guiDescription()
        except AttributeError:
            raise APIException('The type "{}" does not exist'.format(type_osmanager))

        fields = self.add_default_fields(fields, ['name', 'comments', 'tags'])
        return fields


class OSManagerAutoSchema(GUIAutoSchema):
    """Авто схема для CRUD API менеджеров ОС"""

    def get_gui_types(self):
        """Возвращает список зарегистрированных менеджеров ОС"""
        return factory().providers().values()

    def _get_reference(self, serializer):
        return {
            'oneOf': [
                {'$ref': '#/components/schemas/' + gui_type.type()}
                for gui_type in self.get_gui_types()
            ]
        }

    def get_gui_component(self, gui_type, components):

        if self.view.action == 'destroy':
            return {}

        gui_fields = self.map_gui_fields(
            gui_type.guiDescription(),
            deepcopy(components['OSManager']),
        )
        gui_fields['properties']['data_type'] = {
            'type': 'string',
            'enum': [gui_type.type()],
            'writeOnly': True,
            'description': 'Тип объекта',
        }
        gui_fields['required'].append('data_type')

        return gui_fields


class OSManagersCRUD(CrudCommonViewSet):
    """
    CRUD API для менеджеров ОС
    retrieve:
        Получить информацию о параметрах гостевой ОС
    list:
        Получить параметры гостевых ОС из БД
    create:
        Добавить параметры гостевой ОС в БД
    update:
        Обновить параметры гостевых ОС
    destroy:
        Удалить параметры гостевых ОС
    """

    schema = OSManagerAutoSchema()
    serializer_class = OSManagerSerializer
    permission_classes = [AccessPermission]
    lookup_field = 'uuid'
    model = OSManager

    def get_queryset(self):
        available_types = [osmanager.type() for osmanager in factory().providers().values()]
        return self.model.objects.filter(data_type__in=available_types).all()

    @has_role_permission(OSMANAGER_VIEW)
    def retrieve(self, request, *args, **kwargs):
        return super().full_retrieve(request, *args, **kwargs)

    @has_role_permission(OSMANAGER_VIEW)
    def list(self, request, *args, **kwargs):
        return super().list(request, *args, **kwargs)

    @has_role_permission(OSMANAGER_CREATE)
    def create(self, request, *args, **kwargs):
        return super().create(request, *args, **kwargs)

    @has_role_permission(OSMANAGER_CHANGE)
    def update(self, request, *args, **kwargs):
        return super().update(request, *args, **kwargs)

    def perform_destroy(self, instance):
        if instance.deployedServices.count() > 0:
            raise ValidationError(_("Can't delete an OS Manager with services pools associated"))
        super().perform_destroy(instance)

    @has_role_permission(OSMANAGER_DELETE)
    def destroy(self, request, *args, **kwargs):
        return super().destroy(request, *args, **kwargs)


class OSManagersOverview(ListModelMixin, GenericViewSet):
    """
    Получение списка доступных параметров гостевых
    ОС (без дополнительных полей под каждый тип)
    """

    serializer_class = OSManagerOverviewSerializer
    permission_classes = [AccessPermission]

    def get_queryset(self):
        available_types = [osmanager.type() for osmanager in factory().providers().values()]
        return OSManager.objects.filter(data_type__in=available_types).all()

    @has_role_permission(OSMANAGER_VIEW)
    def list(self, request, *args, **kwargs):
        return super().list(request, *args, **kwargs)


class OSManagersTypes(GenericViewSet):
    """
    Получение списка доступных типов для параметров гостевой ОС
    """

    serializer_class = TypeSerializer
    permission_classes = [AccessPermission]

    @has_role_permission(OSMANAGER_VIEW)
    def list(self, request, *args, **kwargs):
        serializer = self.get_serializer(factory().providers().values(), many=True)
        return Response(sorted(serializer.data, key=lambda f: f['name']))
