import logging

from rest_framework import status
from rest_framework.response import Response

from termidesk.core.managers.PolicyManager import PolicyManager
from termidesk.rest.handlers import global_policies_update_event, global_policies_delete_event
from termidesk.api.permissions import AccessPermission, PoliciesPermission
from termidesk.api.webui.serializers.policy import PolicySerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.models.PolicyGlobal import PolicyGlobal
from termidesk.core.policies.utils import registry, get_policy_class_by_type

logger = logging.getLogger(__name__)


class Policies(CommonViewSet):
    """
    gui:
        Спецификация полей GUI для политики конкретного типа.
    list:
        Список всех доступных политик.
    """

    serializer_class = PolicySerializer
    permission_classes = [AccessPermission, PoliciesPermission]

    lookup_field = "data_type"

    def get_queryset(self):
        db_policies = PolicyGlobal.get_policies_by_types()
        code_policies = registry().policies()
        elems = []
        for code_policy in code_policies.values():
            db_policy = db_policies.get(code_policy.typeType)
            if db_policy is None:
                values = code_policy.get_default_values_dict()
            else:
                values = db_policy.getInstance().valuesDict()

            elems.append(values)

        return elems

    def get_policy_object(self):
        data_type = self.kwargs['data_type']
        instance = PolicyGlobal.objects.get(data_type=data_type)
        self.check_object_permissions(self.request, instance)
        return instance

    def get_by_type(self, request, data_type: str):
        """Получить информацию о политике."""

        policy_cls = get_policy_class_by_type(data_type)
        policy_values = PolicyManager().get_global_policy(policy_cls).valuesDict()
        serializer = self.get_serializer(policy_values, context={"request": request})

        return Response(serializer.data)

    def update_by_type(self, request, data_type: str):
        """Обновить параметры политики."""
        data = request.data

        # для отработки процессов в сериализаторе необходимо передать data_type
        data["data_type"] = data_type

        if PolicyGlobal.objects.filter(data_type=data_type).exists():
            instance = self.get_policy_object()
            serializer = self.get_serializer(instance=instance, data=data)
        else:
            self.check_object_permissions(self.request, None)
            serializer = self.get_serializer(data=data)

        serializer.is_valid(raise_exception=True)
        serializer.save()

        global_policies_update_event(
            username=request.user.name,
            authenticator_name=request.user.manager.name,
            policy_name=serializer.instance.name,
            value=serializer.instance.getType().get_view_value(serializer.instance.get_value()),
            authenticator_uuid=request.user.manager.uuid,
            value_raw=serializer.instance.get_value(),
        )
        return Response(serializer.instance.getInstance().valuesDict())

    def delete_by_type(self, request, data_type: str):
        """Сброс глобальной политики"""
        if PolicyGlobal.objects.filter(data_type=data_type).exists():
            instance = self.get_policy_object()

            instance.delete()

            global_policies_delete_event(
                username=request.user.name,
                authenticator_name=request.user.manager.name,
                policy_name=instance.name,
                authenticator_uuid=request.user.manager.uuid,
            )

        return Response(status=status.HTTP_204_NO_CONTENT)

    def get_gui_params(self, **kwargs):
        type_policy = kwargs.get("data_type", "")

        policy = get_policy_class_by_type(type_policy)
        return policy.guiDescription()
