# -*- encoding: utf-8 -*-

from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from ..serializers.providers import ProviderSerializer, ProviderOverviewSerializer
from termidesk.models.UserPermission import (
    PROVIDER_CHANGE, PROVIDER_VIEW, PROVIDER_CREATE, PROVIDER_DELETE
)
from termidesk.api.webui.views.common_viewset import CrudCommonViewSet
from termidesk.notifications.entities.ProviderEventsHelper import ProviderEventsHelper
from termidesk.rest.handlers import set_crud_event_from_viewset, UPDATE_ACTION

from uds.core.util.Config import GlobalConfig
from uds.models import Provider
from uds.core.Environment import Environment
from uds.core.services import factory

# Редакция Termidesk for Astra не содержит поставщика услуг vAir
try:
    from termidesk.services.vair import vair_providers
except ImportError:
    vair_providers = []


from ...decorators import has_role_permission


class Providers(CrudCommonViewSet):
    """
    overview:
        Список всех доступных поставщиков ресурсов
    test:
        Тестирование подключения по переданным данным
    create:
        Создание поставщика ресурсов
        Для просмотра доступных типов поставщиков ресурсов
        обратитесь по следующему пути: /api/webui/providers/types
    retrieve:
        Получить информацию о поставщике ресурсов
    update:
        Обновить поставщика ресурсов
    delete:
        Удалить поставщика ресурсов
    maintenance:
        Обновить параметр техобслуживания поставщика ресурсов
    providers_types:
        Спецификация доступных поставщиков ресурсов
    """

    lookup_field = 'uuid'
    model = Provider
    serializer_class = ProviderSerializer
    serializer_classes = {
        'overview': ProviderOverviewSerializer,
        'retrieve': ProviderOverviewSerializer,
        'maintenance': ProviderOverviewSerializer,
    }

    def get_queryset(self):
        if self.action == 'overview':
            queryset = self.model.objects.filter(data_type__in=[
               provider.typeType for provider in factory().providers().values()])
        else:
            queryset = self.model.objects.all()
        return queryset

    @has_role_permission(PROVIDER_VIEW)
    def overview(self, request, *args, **kwargs):
        return super().list(request, *args, **kwargs)

    def test(self, request, *args, **kwargs):
        type_prov = kwargs.get('type')
        provider = factory().lookup(type_prov)
        params = request.data.copy()
        params['_request'] = request
        res = provider.test(Environment.getTempEnv(), params)

        return Response(self.success() if res[0] is True else res[1])

    @has_role_permission(PROVIDER_CREATE)
    def create(self, request, *args, **kwargs):
        return super().create(request, *args, **kwargs)

    @has_role_permission(PROVIDER_VIEW)
    def retrieve(self, request, *args, **kwargs):
        return super().full_retrieve(request, *args, **kwargs)

    @has_role_permission(PROVIDER_CHANGE)
    def update(self, request, *args, **kwargs):
        return super().update(request, *args, **kwargs)

    def perform_destroy(self, instance):
        if instance.services.count() > 0:
            raise ValidationError(_("Can't delete providers with services already associated"))
        super().perform_destroy(instance)

    @has_role_permission(PROVIDER_DELETE)
    def delete(self, request, *args, **kwargs):
        return super().destroy(request, *args, **kwargs)

    @has_role_permission(PROVIDER_CHANGE)
    def maintenance(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        serializer.maintenance(instance)

        provider_events = ProviderEventsHelper()
        provider_events.maintenance_mode_changed(instance, request.user)

        set_crud_event_from_viewset(self, serializer, UPDATE_ACTION)

        return Response(serializer.data)

    @has_role_permission(PROVIDER_VIEW)
    def providers_types(self, request):
        providers_list = []
        providers = list(factory().providers().values())

        # TSRV-506: Если системная настройка `EXPERIMENTAL_VAIR_PROVIDER_ENABLED` выключена,
        # то необходимо исключить из выдачи провайдеры vAir
        vair_enabled = GlobalConfig.EXPERIMENTAL_VAIR_PROVIDER_ENABLED.getBool(force=True)
        if not vair_enabled:
            providers = \
                [_provider for _provider in providers if _provider not in vair_providers]

        for provider in providers:
            item = {
                'name': provider.name(),
                'type': provider.type(),
                'description': provider.description(),
                'icon': provider.icon(),
            }
            providers_list.append(item)

        return Response(sorted(providers_list, key=lambda f: f["type"]))
