import logging
import uuid

from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from rest_framework.exceptions import APIException
from rest_framework.generics import ListAPIView
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST
from termidesk.api.decorators import has_role_permission
from termidesk.api.webui.serializers.service import ServiceDetailSerializer, ServiceListSerializer
from termidesk.api.webui.serializers.type import TypeSerializer
from termidesk.api.webui.views.common_viewset import ChildCommonViewSet
from termidesk.models.UserPermission import PROVIDER_SERVICE_MANAGE, PROVIDER_VIEW
from termidesk.rest.handlers import CREATE_ACTION, DELETE_ACTION, READ_ACTION, UPDATE_ACTION, set_crud_event
from uds.models.Provider import Provider
from uds.models.Service import Service
from uds.core import services

logger = logging.getLogger(__name__)


class ServicesOverview(ListAPIView):
    """Список всех доступных шаблонов рабочих мест"""
    serializer_class = ServiceListSerializer

    def get_queryset(self):
        provider_uuid = self.kwargs.get('provider_uuid', '')
        provider = get_object_or_404(Provider, uuid=provider_uuid)
        registered_service_types = services.factory().get_service_types()
        return provider.services.filter(data_type__in=registered_service_types)

    @has_role_permission(PROVIDER_VIEW)
    def list(self, request, *args, **kwargs):
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().list(request, *args, **kwargs)


class ServicesType(ListAPIView):
    """Список всех доступных типов шаблонов рабочих мест"""

    serializer_class = TypeSerializer

    @has_role_permission(PROVIDER_VIEW)
    def list(self, request, provider_uuid: uuid):
        provider = get_object_or_404(Provider, uuid=provider_uuid)
        offers = [
            {
                'name': _(type_.name()),
                'type': type_.type(),
                'description': _(type_.description()),
                'icon': type_.icon().replace('\n', '')
            }
            for type_ in provider.getType().getServicesTypes()

            # Запрет отображения шаблона ПК СВ БРЕСТ типа pksvbrestFullLiveService,
            # так как он устарел. Этот запрет будет удален после удаления типа
            # pksvbrestFullLiveService из плагинов в задаче [TSRV-752]
            if type_.type() != 'pksvbrestFullLiveService'
        ]

        serializer = TypeSerializer(data=offers, many=True)

        return Response(sorted(serializer.initial_data, key=lambda f: f['name']))


class ServiceDetail(ChildCommonViewSet):
    model = Service
    serializer_class = ServiceDetailSerializer

    def get_serializer(self, *args, **kwargs):
        kwargs['context'] = self.get_serializer_context()

        provider_uuid = self.kwargs.get('provider_uuid')
        kwargs['provider'] = (get_object_or_404(Provider, uuid=provider_uuid)
                              if provider_uuid is not None else '')
        return self.serializer_class(*args, **kwargs)

    @has_role_permission(PROVIDER_VIEW)
    def retrieve(self, request, provider_uuid: uuid, uuid: uuid):
        """
        Получить информацию о шаблоне рабочих мест.
        """
        provider = get_object_or_404(Provider, uuid=provider_uuid)
        service = get_object_or_404(Service, uuid=uuid, provider__uuid=provider.uuid)
        serializer = self.get_serializer(service, context={'request': request})

        set_crud_event(
            self.request.user.manager.name,
            self.request.user.name,
            self.request.ip,
            self.model.__name__,
            service.uuid,
            service.data_type,
            service.name,
            READ_ACTION
        )
        params = serializer.data
        return Response(self.fill_intance_fields(service, params))

    @has_role_permission(PROVIDER_SERVICE_MANAGE)
    def create(self, request, *args, **kwargs):
        """ Создать шаблон рабочих мест. """
        return super().create(request, *args, **kwargs)

    def perform_create(self, serializer):
        serializer.save()

        set_crud_event(
            self.request.user.manager.name,
            self.request.user.name,
            self.request.ip,
            self.model.__name__,
            serializer.instance.uuid,
            serializer.instance.data_type,
            serializer.instance.name,
            CREATE_ACTION
        )
        self._set_read_instance_audit(serializer.instance)
        self._set_update_parents_audit(serializer.instance.provider)

    @has_role_permission(PROVIDER_SERVICE_MANAGE)
    def update(self, request, provider_uuid: uuid, uuid: uuid):
        """ Обновить шаблон рабочих мест. """
        provider = get_object_or_404(Provider, uuid=provider_uuid)
        service = get_object_or_404(Service, uuid=uuid, provider__uuid=provider.uuid)
        serializer = self.get_serializer(instance=service, data=request.data, context={'request': request})

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            service.uuid,
            service.data_type,
            service.name,
            UPDATE_ACTION
        )

        self._set_read_instance_audit(service)
        self._set_update_parents_audit(provider)

        return Response(serializer.data)

    @has_role_permission(PROVIDER_SERVICE_MANAGE)
    def delete(self, request, provider_uuid: uuid, uuid: uuid):
        """ Удалить шаблон рабочих мест. """
        provider = get_object_or_404(Provider, uuid=provider_uuid)
        service = get_object_or_404(Service, uuid=uuid, provider__uuid=provider.uuid)

        try:
            if service.deployedServices.count() == 0:
                service.delete()

                set_crud_event(
                    request.user.manager.name,
                    request.user.name,
                    request.ip,
                    service.__class__.__name__,
                    service.uuid,
                    service.data_type,
                    service.name,
                    DELETE_ACTION
                )
                self._set_update_parents_audit(provider)

                return Response('deleted')

        except Exception as error:
            logger.exception('Deleting service')
            return Response(error, status=HTTP_400_BAD_REQUEST)

        raise APIException('Item has associated deployed services')
