from django.shortcuts import get_object_or_404
from rest_framework.response import Response

from termidesk.api.decorators import has_object_role_permission
from termidesk.api.webui.serializers.servicespool.message import SendMessageSerializer
from termidesk.api.webui.schemas import SendMessageSchema
from termidesk.api.webui.views.common_viewset import ChildCommonViewSet
from termidesk.models.UserPermission import DS_USER_SERVICE_MANAGE
from termidesk.notifications.delayed_task.message_userservice import SUGGESTED_TIME, MessageUserService, TAG
from termidesk.rest.handlers import user_workplace_message_sent_event
from uds.core.util import log
from uds.core.util.State import State
from uds.core.jobs.DelayedTaskRunner import DelayedTaskRunner
from uds.models.ServicesPool import ServicePool
from uds.REST.model import OK


class SendMessageAssignedUserService(ChildCommonViewSet):
    """Класс обработки запросов для отправления сообщений назначенным рабочим местам
    get_assigned_message: Отправка сообщения назначенным рабочим местам
    """
    model = ServicePool
    serializer_class = SendMessageSerializer
    schema = SendMessageSchema(operation_id_base='WebuiServicepoolSendMessage')

    # минимальная версия агента для отправки сообщений
    MIN_VERSION_AGENT = (3, 2, 0)

    @has_object_role_permission(DS_USER_SERVICE_MANAGE, ServicePool, "uuid")
    def get_assigned_message(self, request, uuid):
        """ Отправка сообщения назначенным рабочим местам"""
        message = request.data
        workplace = get_object_or_404(self.model, uuid=uuid)

        # Валидация сообщения
        serializer = SendMessageSerializer(data=request.data, context={'request': request})
        serializer.is_valid(raise_exception=True)

        for user_service in workplace.assignedUserServices().filter(state=State.USABLE):
            tag = TAG + user_service.uuid
            if DelayedTaskRunner.runner().checkExists(tag):
                log_message = 'Предыдущее сообщение находится в статусе отправления'
                log.doLog(user_service, log.ERROR, log_message, log.SERVICE)
                continue

            log_message = 'User {}({}) sent "{}" message: "{}".'.format(
                request.user.name,
                request.user.manager.name,
                message['level'],
                message['message']
            )
            log.doLog(user_service, log.INFO, log_message, log.SERVICE)
            delayed_task = MessageUserService(
                user_service.uuid,
                message['message'],
                message['level']
            )
            delayed_task.register(SUGGESTED_TIME, tag)

            user_workplace_message_sent_event(
               authenticator_name=request.user.manager.name,
               authenticator_uuid=request.user.manager.uuid,
               username=request.user.name,
               deployed_service_name=workplace.name,
               deployed_service_uuid=workplace.uuid,
               user_service_name=user_service.getName(),
               user_service_uuid=user_service.uuid,
               msg_level=message['level'],
               msg_text=message['message']
            )

        return Response(OK)
