import logging
import uuid

from django.shortcuts import get_object_or_404
from rest_framework import status
from rest_framework.response import Response

from termidesk.core.managers.PolicyManager import PolicyManager
from termidesk.rest.handlers import policies_update_event, policies_delete_event
from termidesk.api.permissions import AccessPermission, PoliciesPermission
from termidesk.api.webui.serializers.servicespool.policy import (DEPLOYED_SERVICE_ID,
                                                                 PolicyDeployedServiceSerializer)
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.api.webui.views.servicespool import SERVICE_POOL_UUID
from termidesk.core.policies.utils import get_policy_class_by_type, registry
from termidesk.models.PolicyDeployedService import PolicyDeployedService
from termidesk.models.PolicyGlobal import PolicyGlobal
from uds.models.ServicesPool import DeployedService

logger = logging.getLogger(__name__)


class PoliciesDeployedServices(CommonViewSet):
    """
    list:
        Список всех политик фонда рабочих мест.
    """

    serializer_class = PolicyDeployedServiceSerializer
    permission_classes = [AccessPermission, PoliciesPermission]

    lookup_field = "data_type"

    deployed_service = None

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)

        servicespool_uuid = kwargs.get(SERVICE_POOL_UUID)
        self.deployed_service = get_object_or_404(DeployedService, uuid=servicespool_uuid)

    def get_queryset(self):
        deployed_service_policies = PolicyDeployedService.get_policies_by_types(self.deployed_service.id)
        global_policies = PolicyGlobal.get_policies_by_types()
        code_policies = registry().policies()
        elems = []
        for code_policy in code_policies.values():
            db_policy = deployed_service_policies.get(
                code_policy.typeType, global_policies.get(code_policy.typeType)
            )
            if db_policy is None:
                values = code_policy.get_default_values_dict()
                values['view_value'] += ' (унаследовано)'
            else:
                values = db_policy.getInstance().valuesDict()
            elems.append(values)

        return elems

    def get_policy_object(self):
        data_type = self.kwargs['data_type']
        instance = PolicyDeployedService.get_policy(data_type, self.deployed_service.id)
        self.check_object_permissions(self.request, instance)
        return instance

    def get_serializer(self, *args, **kwargs):
        kwargs['context'] = self.get_serializer_context()
        kwargs[DEPLOYED_SERVICE_ID] = self.deployed_service.id if self.deployed_service is not None else None

        return self.serializer_class(*args, **kwargs)

    def get_by_type(self, request, servicespool_uuid: uuid, data_type: str):
        """Получить информацию о политиках фонда рабочих мест."""
        policy_cls = get_policy_class_by_type(data_type)
        policy = PolicyManager().get_deployed_service_policy(policy_cls, self.deployed_service)
        policy_values = policy.valuesDict()
        serializer = self.get_serializer(policy_values, context={"request": request})

        return Response(serializer.data)

    def update_by_type(self, request, servicespool_uuid: uuid, data_type: str):
        """ Обновить параметры политики фонда рабочих мест. """
        data = request.data
        # для отработки процессов в сериализаторе необходимо передать data_type
        data["data_type"] = data_type

        instance = None
        if PolicyDeployedService.is_policy_exists(data_type, self.deployed_service.id):
            instance = self.get_policy_object()

        self.check_object_permissions(request, instance)

        serializer = self.get_serializer(instance=instance, data=data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        policies_update_event(
            username=request.user.name,
            authenticator_name=request.user.manager.name,
            policy_name=serializer.instance.name,
            deployed_service_name=self.deployed_service.name,
            value=serializer.instance.getType().get_view_value(serializer.instance.get_value()),
            authenticator_uuid=request.user.manager.uuid,
            deployed_service_uuid=self.deployed_service.uuid,
            value_raw=serializer.instance.get_value(),
        )

        return Response(serializer.instance.getInstance().valuesDict())

    def delete_by_type(self, request, servicespool_uuid: uuid, data_type: str):
        """Сброс политики фонда рабочих мест"""
        if PolicyDeployedService.is_policy_exists(data_type, self.deployed_service.id):
            instance = self.get_policy_object()

            instance.delete()

            policies_delete_event(
                username=request.user.name,
                authenticator_name=request.user.manager.name,
                policy_name=instance.name,
                deployed_service_name=self.deployed_service.name,
                authenticator_uuid=request.user.manager.uuid,
                deployed_service_uuid=self.deployed_service.uuid,
            )

        return Response(status=status.HTTP_204_NO_CONTENT)

    def get_gui_params(self, **kwargs):
        type_policy = kwargs.get("data_type", "")

        policy = get_policy_class_by_type(type_policy)
        return policy.guiDescription()
