import logging
import uuid

from django.shortcuts import get_object_or_404
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.status import HTTP_200_OK
from rest_framework.viewsets import ModelViewSet
from termidesk.api.decorators import has_object_role_permission, has_role_permission
from termidesk.api.webui.serializers.servicespool.publication import (PublicationListSerializer,
                                                                      PublicationSerializer)
from termidesk.api.webui.views.common_viewset import ChildCommonViewSet
from termidesk.api.webui.views.servicespool import SERVICE_POOL_UUID
from termidesk.models.UserPermission import DEPLOYED_SERVICE_VIEW, DS_PUBLICATION_MANAGE
from uds.models import DeployedServicePublication
from uds.models.ServicesPool import ServicePool
from uds.models.ServicesPoolPublication import ServicePoolPublication

logger = logging.getLogger(__name__)


class PublicationsOverviewServicePool(ModelViewSet):
    """
    list:
        Список всех доступных публикаций фонда рабочих мест.
    """
    model = ServicePoolPublication
    serializer_class = PublicationListSerializer

    def get_queryset(self):
        service_pool_uuid = str(self.kwargs.get(SERVICE_POOL_UUID, '')).lower()

        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        return service_pool.publications.all()

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def list(self, request, *args, **kwargs):
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().list(request, *args, **kwargs)


class PublicationsServicePool(ChildCommonViewSet):
    """
    gui:
        Спецификация полей интерфейса для публикации фонда рабочих мест.
    cancel:
        Отмена публикации фонда рабочих мест.
    """
    serializer_class = PublicationSerializer

    def get_gui_params(self, **kwargs):
        service_pool_uuid = str(kwargs.get(SERVICE_POOL_UUID, '')).lower()
        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        service = service_pool.service
        service_inst = service.getInstance()

        publication_class = service_inst.publicationType
        if publication_class is None:
            # Для шаблона не предусмотрена публикация
            return []

        gui_class = publication_class.gui
        gui_inst = gui_class(parent=service_pool)
        gui = gui_class.guiDescription(gui_inst)

        return gui

    @has_object_role_permission(DS_PUBLICATION_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def cancel(self, request, servicespool_uuid: uuid, uuid: uuid):
        """
        Invoked to cancel a running publication
        Double invocation (this means, invoking cancel twice) will mean that is a "forced cancelation"
        """
        service_pool_uuid = str(servicespool_uuid).lower()
        uuid = str(uuid).lower()

        if not any((request.user.is_superuser, request.user.is_staff,)):
            logger.debug('Management Permission failed. User {} is not admin or staff member.'
                         .format(request.user))
            self.access_denied()

        deployed_service = get_object_or_404(DeployedServicePublication,
                                             uuid=uuid, deployed_service__uuid=service_pool_uuid)
        deployed_service.cancel()

        return Response(self.success())

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def gui(self, request, *args, **kwargs):
        return super().gui(request, *args, **kwargs)


class PublishSchema(AutoSchema):
    def get_operation(self, path, method):
        operation = super().get_operation(path, method)

        from uds.core import services
        providers = services.factory().providers()
        pub_gui_types = set()
        for prov in list(providers.values()):
            for srv in prov.offers:
                if srv.publicationType is not None:
                    pub_gui_types.add(srv.publicationType.gui)

        pub_gui_types = sorted(pub_gui_types, key=lambda clazz: clazz.__name__)
        for gui in pub_gui_types:
            gui_desc = gui.guiDescription()
            props = {}
            for field in gui_desc:
                props[field["name"]] = {
                    "type": "string"
                }
            description_tpl = "Набор параметров для публикаций c интерфейсом {pub_type}"
            description = description_tpl.format(pub_type=gui.__name__)
            tpl = {
                "name": gui.__name__,
                "in": "query",
                "required": False,
                "description": description,
                "schema": {
                    "type": "object",
                    "properties": props,
                },
                "additionalProperties": False,
            }
            operation["parameters"].append(tpl)

        operation['parameters'].append({
            "name": "tags",
            "in": "query",
            "required": False,
            "description": "Теги",
            'schema': {'type': 'list'}
        })
        return operation


class PublishServicePool(ChildCommonViewSet):
    """
    publish:
        Публикация фонда рабочих мест.
    """
    serializer_class = PublicationSerializer
    schema = PublishSchema()

    @has_object_role_permission(DS_PUBLICATION_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def publish(self, request, servicespool_uuid: uuid):
        """
        Custom method "publish", provided to initiate a publication of a deployed service
        """
        service_pool = get_object_or_404(ServicePool, uuid=str(servicespool_uuid).lower())

        validated_data = self._validate_publish_data(service_pool, self.request.query_params)

        changelog = validated_data["changelog"]
        # Параметр changelog хранится в отдельном поле модели
        # по историческим причинам
        del(validated_data["changelog"])

        if not any((request.user.is_superuser, request.user.is_staff,)):
            logger.debug('Management Permission failed. User {} is not admin or staff member.'
                         .format(request.user))
            self.access_denied()

        logger.debug('Custom "publish" invoked for {}'.format(service_pool))
        # Can raise exceptions that will be processed on response
        service_pool.publish(changelog, validated_data)

        self._set_update_parents_audit(service_pool)

        return Response(self.success(), status=HTTP_200_OK)

    def _validate_publish_data(self, service_pool, query_params):
        # В дальнейшем данный метод нужно будет перенести в сериализатор,
        # если он будет реализован.
        validated_data = {}

        service = service_pool.service
        service_inst = service.getInstance()

        publication_class = service_inst.publicationType
        if publication_class is None:
            # Для шаблона не предусмотрена публикация
            # По идее мы здесь быть не должны из-за других проверок
            raise Exception("Publication class not specified for {}".format(service.data_type))

        # Простая валидация списка обязательных параметров
        gui_spec = publication_class.gui.guiDescription()

        for field in gui_spec:
            val = query_params.get(field['name'], None)
            # Для обязательных неуказанных параметров вызываем ошибку API
            if val is None and field["gui"]["required"]:
                raise ValidationError("Missing parameter {}".format(field['name']))

            # Для необязательных неуказанных параметров проставляем значение по-умолчанию
            if val is None:
                val = field["gui"]["defvalue"]

            validated_data[field['name']] = val

        return validated_data
