"""События, связанные с веб-интерфейсом"""

from __future__ import unicode_literals

from collections import namedtuple

from .base import BaseEventMixin, build_name


__all__ = (
    'WEB_USER_LOGIN_V1',
    'WEB_USER_LOGOUT_V1',
    'WEB_ENTITY_ACTION',
    'WEB_SYSTEM_CONFIG_CHANGED_V1',
    'WEB_LICENSE_UPDATED_V1',
    'WEB_LOGOFF_SESSION_USERSERVICE',
    'WEB_DISCONNECT_SESSION_USERSERVICE',
)


class WEB_USER_LOGIN_V1(
        # Наследуем общую для всех классов событий функциональность
        BaseEventMixin,
        # Создаем родительский класс на базе кортежа с нужными аттрибутами
        namedtuple(
            # Имя родительского типа именованого кортежа
            'WEB_USER_LOGIN_V1',
            # Определение аттрибутов события
            ['authenticator', 'username', 'ip'])):
    '''
       Событие: Вход пользователя в систему через веб-интерфейс
       Данные:
         :param authenticator: Название домена аутентифкации пользователя - args[0]
         :param username: Логин пользователя - args[1]
         :param ip: IP адрес, с которого был сделан запрос - args[2]
    '''
    __slots__ = ()
    _NAME = build_name('web.UserLogin')  # Задаем имя события
    _VERSION = 1  # Версия схемы события
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" вошел в систему с ip-адреса {args[2]}'


class WEB_USER_LOGOUT_V1(
        BaseEventMixin,
        namedtuple(
            'WEB_USER_LOGOUT_V1',
            ['authenticator', 'username', 'ip'])):
    '''
       Событие: Выход пользователя из веб-интерфейса (logout)
       Данные:
         :param authenticator: Название домена аутентифкации пользователя - args[0]
         :param username: Логин пользователя - args[1]
         :param ip: IP адрес, с которого был сделан запрос - args[2]
    '''
    __slots__ = ()
    _NAME = build_name('web.UserLogout')
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" вышел из системы (ip-адрес {args[2]})'


class WEB_SYSTEM_CONFIG_CHANGED_V1(
    BaseEventMixin,
    namedtuple(
        'WEB_SYSTEM_CONFIG_CHANGED_V1',
        ['authenticator', 'username', 'ip', ])):
    """
        Событие: Изменение системных параметров Диспетчера подключений
        Данные:
            :param authenticator: Название домена аутентификации пользователя - args[0]
            :param username: Логин пользователя - args[1]
            :param ip: IP адрес, с которого был сделан запрос - args[2]
    """
    __slots__ = ()
    _NAME = build_name('web.SystemConfigChanged')
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" изменил системные параметры ' \
                           '(ip-адрес {args[2]})'


class WEB_ENTITY_ACTION(
    BaseEventMixin,
    namedtuple(
        'WEB_CRUD_ENTITY_V1',
        ['authenticator', 'username', 'ip', 'entity', 'uuid', 'subtype', 'name', 'action'])):
    """Событие: CRUD операции с объектами через REST API.

       Данные:
       :param authenticator: название домена аутентифкации пользователя - args[0]
       :param username: логин пользователя - args[1]
       :param ip: IP адрес, с которого был сделан запрос - args[2]
       :param entity: тип сущности - args[3]
       :param uuid: уникальный идентификатор - args[4]
       :param subtype: тип объекта - args[5]
       :param name: название объекта - args[6]
       :param action: действие над объектом (CRUD) - args[7]
    """
    _NAME = build_name('web.EntityAction')
    __slots__ = ()
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" выполнил операцию {args[7]} для объекта ' \
                           '{args[3]} ({args[4]}) {args[5]} "{args[6]}" ' \
                           '(ip-адрес {args[2]})'


class WEB_LICENSE_UPDATED_V1(
    BaseEventMixin,
    namedtuple(
        'WEB_LICENSE_UPDATED_V1',
        ['authenticator', 'username', 'ip', 'license_file_name'])):
    """Событие: Загрузка файла лицензии через REST API.

       Данные:
       :param authenticator: название домена аутентификации пользователя - args[0]
       :param username: логин пользователя - args[1]
       :param ip: IP адрес, с которого был сделан запрос - args[2]
       :param license_file_name: имя файла лицензии - args[3]
    """
    _NAME = build_name('web.LicenseUpdated')
    __slots__ = ()
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" загрузил новый файл лицензии {args[3]} ' \
                           'с ip-адреса {args[2]}'


class WEB_LOGOFF_SESSION_USERSERVICE(
    BaseEventMixin,
    namedtuple(
        'WEB_LOGOFF_SESSION_USERSERVICE',
        ['user', 'userservice'])):
    """Событие: Прекращение сессии пользователя по команде с сервера.

       Данные:
       :param user: логин пользователя - args[0]
       :param userservice: данные гостевой ВМ, сессию которой прекратили - args[1]
    """
    _NAME = build_name('web.LogoffUserservice')
    __slots__ = ()
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[0]}" отправил запрос на прекращение сессии {args[1]}'


class WEB_DISCONNECT_SESSION_USERSERVICE(
    BaseEventMixin,
    namedtuple(
        'WEB_DISCONNECT_SESSION_USERSERVICE',
        ['user', 'userservice'])):
    """Событие: Сброс сессии пользователя по команде с сервера.

       Данные:
       :param user: логин пользователя - args[0]
       :param userservice: данные гостевой ВМ, сессию которой сбросили - args[1]
    """
    _NAME = build_name('web.DisconnectUserservice')
    __slots__ = ()
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[0]}" отправил запрос на сброс сессии {args[1]}'
