import json
import logging
from abc import ABCMeta, abstractmethod

from uds.core import Module
from uds.core.ui.UserInterface import gui

logger = logging.getLogger()


class BasePolicy(Module):
    """Базовый класс политик"""

    __metaclass__ = ABCMeta
    __ui_version__ = 2

    typeName = "Base PolicyManager"
    typeType = "BasePolicy"
    typeDescription = "Base Policy"

    # Допустимые значения в виде словаря,
    # ключи - все допустимые значения,
    # значения - описание для отображения в интерфейсе.
    ALLOW_VALUES = {}

    _value = ""

    description = gui.TextField(
        label="Описание",
        required=False,
        rdonly=True,
        multiline=4,
        order=1
    )

    value = gui.InputField(
        label="Значение",
        required=True,
        defvalue=0,
        order=2
    )

    JSON_DATA_VERSION_1 = 1

    def __init__(self, environment, values):
        super().__init__(environment, values)
        if values is not None:
            self.set_value(values.get("value"))
        else:
            self.set_value(self.value.defValue)

    def marshal(self):
        data = {
            'version': self.JSON_DATA_VERSION_1,
            'value': self._value,
        }

        return json.dumps(data)

    def unmarshal(self, str_):
        if not isinstance(str_, str):
            str_ = str_.decode("utf-8")
        data = json.loads(str_)
        if data['version'] == self.JSON_DATA_VERSION_1:
            self.set_value(data.get('value'))
            logger.debug("Value: {value}".format(value=data['value']))
        else:
            raise Exception("Unsupported serialized data version {}".format(data['version']))

    def set_value(self, value):
        self._value = value

    def get_value(self):
        return self._value

    def valuesDict(self):
        value = self.get_value()
        return PolicyDTO.init(type(self), value).get_values()

    @classmethod
    def get_default_values_dict(cls):
        """Возвращает словарь значений по-умолчанию"""
        return PolicyDTO.init(cls).get_values()

    @classmethod
    @abstractmethod
    def get_view_value(cls, value):
        """Получение текстового представления значения, для отображения в интерфейсе."""
        return cls.ALLOW_VALUES.get(value)

    @classmethod
    @abstractmethod
    def check_value(cls, value):
        """Проверка переданного значения."""
        if cls.ALLOW_VALUES.get(value) is None:
            raise ValueError('Передано не допустимое значение "value"')


class PolicyDTO:
    """
    Контейнер для хранения структуры полей политики.

    Этот класс в будущем можно переделать на dataclass(>=python3.7).
    """

    __value = None

    __name = None
    __data_type = None
    __view_value = None

    def __init__(self, name, description, data_type, value, view_value):
        self.__name = name
        self.__description = description
        self.__data_type = data_type
        self.__value = value
        self.__view_value = view_value

    def get_values(self):
        return {
            "name": self.__name,
            "description": self.__description,
            "data_type": self.__data_type,
            "value": self.__value,
            "view_value": self.__view_value,
        }

    @classmethod
    def init(cls, policy_cls, value=None):
        """Инициирует DTO класс на основе класса политики"""

        value = value or policy_cls.value.defValue
        return cls(
            policy_cls.typeName,
            policy_cls.typeDescription,
            policy_cls.typeType,
            value,
            policy_cls.get_view_value(value)
        )
