import weakref
from operator import itemgetter

from guardian.exceptions import MixedContentTypeError, WrongAppError
from guardian.shortcuts import (
    _get_queryset,
    _handle_pk_field,
    get_content_type,
    get_group_obj_perms_model,
    groupby,
)

from django.contrib.contenttypes.models import ContentType


def get_objects_for_groups(groups, perms, klass=None, any_perm=False):
    """
    Returns queryset of objects for which a given ``groups`` has *all*
    permissions present at ``perms``.

    Copied from method guardian.shortcuts.get_objects_for_group
    without processing code of param "accept_global_perms".
    """

    builder = ObjectsForGroupsQuerysetBuilder(
        groups=groups,
        perms=perms,
        klass=klass,
        any_perm=any_perm,
    )
    return builder.get_queryset()


class ObjectsForGroupsQuerysetBuilder:
    """
    Вспомогательный класс, содержащий в себе логику
    для функции get_objects_for_groups.

    Публичный метод get_queryset возвращает QuerySet объектов,
    для которых данные группы groups имеют все разрешения, указанные в perms.

    В perms должны содержаться строки, которые соответствуют разрешениям группы.
    Нельзя одновременно использовать разрешения из разных приложений.
    Правильно: perms = ['app1.add_model1', 'app1.add_model2', 'app1.view_model2']
    Неправильно: perms = ['app1.add_model1', 'app2.add_model2']

    :param: klass - Класс модели, используется, если не удается автоматически
    определить content_type объектов через perms

    :param: any_perms - поиск по любым разрешениям
    """

    def __init__(self, groups, perms, klass=None, any_perm=False):
        if isinstance(perms, str):
            perms = [perms]

        self.groups = groups
        self.perms = [ObjectsForGroupsPermEntry(perm) for perm in perms]
        self.klass = klass
        self.any_perm = any_perm
        self.content_type = None
        self.app_label = None
        self._validator = ObjectsForGroupsItemsValidator(self)

    def get_queryset(self):
        """Возвращает QuerySet с объектами"""

        codenames = self._compute_codenames()
        self._set_app_label()
        self._set_content_type(codenames)

        return self._get_queryset(codenames)

    def _compute_codenames(self):
        """Возвращает множество строк codenames"""
        return {
            perm.codename
            for perm in self.perms
        }

    def _set_app_label(self):
        """Определяет app_label и проверяет, что все
        perms соответствуют одному app_label.
        """

        app_label = None

        for perm in self.perms:
            app_label = self._validator.validate_perm_app_label(perm, app_label)

        self.app_label = app_label

    def _set_content_type(self, codenames):
        """Определяет content_types и проверяет, что все
        codenames соответствуют одному content_type.
        """
        content_type = None

        for codename in codenames:
            content_type = self._validator.validate_codename_content_type(
                codename,
                content_type,
            )

        if content_type is None and self.klass is not None:
            queryset = _get_queryset(self.klass)
            content_type = get_content_type(queryset.model)

        if content_type is None:
            raise WrongAppError("Cannot determine content type")

        self.content_type = content_type

    def _get_queryset(self, codenames):
        """Генерирует запрос, по результату которого
        будет происходить фильтрация объектов
        """
        qs_helper = ObjectsForGroupsQuerysetHelper(self)

        queryset = qs_helper.get_base_queryset()

        if not self.any_perm and len(codenames) != 0:
            pk_list = qs_helper.get_values_by_codenames_in_group(codenames)
        else:
            pk_list = qs_helper.get_values_by_group_annotation(codenames)

        return queryset.filter(pk__in=pk_list)


class ObjectsForGroupsPermEntry:
    """
    Высокоуровневое представление строки perm
    для ObjectsForGroupsQuerysetBuilder
    """

    def __init__(self, perm):
        self._perm = perm

    @property
    def app_label(self):
        """Извлекает app_label из строки perm"""
        if '.' in self._perm:
            return self._perm.split('.', 1)[0]

        return None

    @property
    def codename(self):
        """Извлекает codename из строки perm"""
        if '.' in self._perm:
            return self._perm.split('.', 1)[1]

        return self._perm


class ObjectsForGroupsItemsValidator:
    """Вспомогательный класс-валидатор для ObjectsForGroupsQuerysetBuilder"""

    def __init__(self, builder):
        self._get_builder = weakref.ref(builder)

    def validate_codename_content_type(self, codename, actual_content_type=None):
        """Проверяет, что текущий codename соответствует
        тому же content_type, что и предыдущие.
        Возвращает content_type
        """

        builder = self._get_builder()

        if builder.app_label is None:
            return None

        content_type = ContentType.objects.get(
            app_label=builder.app_label,
            permission__codename=codename,
        )
        if actual_content_type is not None and actual_content_type != content_type:
            error = 'ContentType was once computed to be {} and another one {}'
            error = error.format(actual_content_type, content_type)
            raise MixedContentTypeError(error)

        return content_type

    @staticmethod
    def validate_perm_app_label(perm, actual_app_label):
        """Проверяет, что у разных разрешений одинаковый параметр app_label.
        Возвращает app_data
        """

        app_label = perm.app_label

        if actual_app_label is not None and actual_app_label != app_label:
            error = 'Given perms must have same app label ({} != {})'
            error = error.format(actual_app_label, app_label)
            raise MixedContentTypeError(error)

        return app_label


class ObjectsForGroupsQuerysetHelper:
    """
    Вспомогательный класс для построения QuerySet
    в ObjectsForGroupsQuerysetBuilder
    """

    def __init__(self, builder):
        self._get_builder = weakref.ref(builder)
        self._base_qs = None
        self._group_model = None

    def get_base_queryset(self):
        """Пытается определить базовый QuerySet для поиска объектов"""
        if self._base_qs is None:

            builder = self._get_builder()

            if builder.klass is None:
                queryset = _get_queryset(builder.content_type.model_class())
            else:
                queryset = _get_queryset(builder.klass)
                if builder.content_type.model_class() != queryset.model:
                    raise MixedContentTypeError(
                        'Content type for given perms and klass differs'
                    )
            self._base_qs = queryset

        return self._base_qs

    def get_groups_perms_queryset(self, codenames):
        """Возвращает QuerySet для групп, для которых
        существуют разрешения из списка codenames
        """
        builder = self._get_builder()

        groups_obj_perms_queryset = (
            self.group_model.objects
            .filter(group__in=[group.id for group in builder.groups])
            .filter(permission__content_type=builder.content_type)
        )
        if len(codenames) != 0:
            groups_obj_perms_queryset = groups_obj_perms_queryset.filter(
                permission__codename__in=codenames,
            )
        return groups_obj_perms_queryset

    @property
    def group_model(self):
        """Возвращает модель группы на основе базового QuerySet"""
        if self._group_model is None:
            queryset = self.get_base_queryset()
            self._group_model = get_group_obj_perms_model(queryset.model)

        return self._group_model

    @property
    def object_pk_field(self):
        """Возвращает поле, по которому будет производиться построение запроса"""

        if self.group_model.objects.is_generic():
            return 'object_pk'

        return 'content_object__pk'

    def get_values_by_codenames_in_group(self, codenames):
        """Возвращает список pk объектов, найденный
        по codenames через разрешения группы
        """
        builder = self._get_builder()
        groups_perms_queryset = self.get_groups_perms_queryset(codenames)

        groups_obj_perms = groups_perms_queryset.values_list(
            self.object_pk_field,
            'permission__codename',
        )
        groups_obj_perms = list(groups_obj_perms)

        # sorting/grouping by pk (first item in a resulting tuple)
        sort_by_pk = itemgetter(0)

        groups_obj_perms = sorted(groups_obj_perms, key=sort_by_pk)
        pk_list = []
        for obj_pk, group in groupby(groups_obj_perms, sort_by_pk):
            obj_codenames = {e[1] for e in group}
            if builder.any_perm or codenames.issubset(obj_codenames):
                pk_list.append(obj_pk)

        return pk_list

    def get_values_by_group_annotation(self, codenames):
        """Возвращает список pk объектов, найденный
        через поле из аннотации групп
        """
        queryset = self.get_base_queryset()
        values = self.get_groups_perms_queryset(codenames)
        object_pk_field = self.object_pk_field

        handle_pk_field = _handle_pk_field(queryset)
        if handle_pk_field is not None:
            values = values.annotate(obj_pk=handle_pk_field(expression=object_pk_field))
            object_pk_field = 'obj_pk'

        return list(values.values_list(object_pk_field, flat=True))
