"""Модуль содержит логику для кэширования данных"""
# -*- encoding: utf-8 -*-

from datetime import datetime, timedelta
from functools import lru_cache, wraps

from django.utils import timezone


def timed_lru_cache(seconds=60, maxsize=128):
    """
    Декоратор для ``lru_cache`` ограничивающий время хранения результата,
    декорируемой с помощью ``lru_cache`` функции

    При декорировании функции необходимо ставить круглые скобки после декоратора
    ::
        @timed_lru_cache()
        def decorated_func():
            pass

    :param seconds: время хранения результата декорируемой функции в секундах
    :type seconds: int
    :param maxsize: максимальный размер для хранения результатов
    :type maxsize: int
    """
    def wrapper_cache(func):
        """Функция-обертка обеспечивает возможность учета времени хранения результатов"""
        func = lru_cache(maxsize=maxsize)(func)
        func.lifetime = timedelta(seconds=seconds)
        func.expiration = datetime.now(tz=timezone.utc) + func.lifetime

        @wraps(func)
        def wrapped_func(*args, **kwargs):
            """Очистка устаревших результатов перед выполнением логики ``lru_cache``"""
            if datetime.now(tz=timezone.utc) >= func.expiration:
                func.cache_clear()
                func.expiration = datetime.now(tz=timezone.utc) + func.lifetime

            return func(*args, **kwargs)

        wrapped_func.cache_info = func.cache_info
        wrapped_func.cache_clear = func.cache_clear

        return wrapped_func

    return wrapper_cache
