import logging
from django.utils.deprecation import MiddlewareMixin
from ipware.ip import get_ip
from termidesk.models.AccessDevice import AccessDevice
from uds.core.util.Config import GlobalConfig
from dateutil import parser

logger = logging.getLogger(__name__)

NGINX = 'nginx'

SUPPORTED_HEADER_FORMATS = (NGINX,)


class DeviceAccessMiddleware(MiddlewareMixin):
    """ Фиксация параметров из https запроса. """

    def __call__(self, request):
        # Проверка доступа к 'Устройства доступа'.
        device_auth_enabled = GlobalConfig.EXPERIMENTAL_DEVICE_AUTH_ENABLED.getBool(True)
        logger.debug('Checking middleware device auth enabled: "%s"', device_auth_enabled)
        if not device_auth_enabled:
            return self.get_response(request)

        cert_missing = request.META.get("HTTP_X_TDSK_SSL_CLIENT_CERT") is None
        if cert_missing:
            # Данные о клиентском сертификате отсутствуют в нужном заголовке
            # Возможно MTLS некорректно настроен на вебсервере или запрос через HTTP
            # Обработку не проводим
            return self.get_response(request)

        dialect = request.META.get("HTTP_X_TDSK_SSL_CLIENT_FORMAT")
        if dialect not in SUPPORTED_HEADER_FORMATS:
            logger.warning(
                ("Unsupported or missing ssl client header format."
                 "HTTP_X_TDSK_SSL_CLIENT_FORMAT=%s"), dialect)
            return self.get_response(request)

        try:
            if dialect == NGINX:
                params = self._parse_nginx_params(request)

            # TODO  Добавить поддержку заголовков Apache

            device, created = AccessDevice.objects.update_or_create(
                certificate_subject=params['subject'],
                defaults={
                    'certificate': params['cert'],
                    'certificate_valid': params["is_valid"],
                    'certificate_subject': params['subject'],
                    'certificate_validity_start': params["date_start"].isoformat(),
                    'certificate_validity_end': params["date_end"].isoformat(),
                    'os': params['os'],
                    'last_seen_ip': params['ip'],
                    # 'access': 'wait'
                }
            )

            # При обновлении записи, статус доступа не изменяется.
            # При необходимости обнулять статус доступа после обновления,
            # необходимо удалить условие ниже и раскомментировать строку
            # 'access': 'wait' в словаре defaults и обновить тесты
            if created:
                device.access = 'wait'
                device.save()

        except Exception as error:
            logger.exception(error)

        return self.get_response(request)

    def _parse_nginx_params(self, request):
        # Обработка заголовков с синтаксисом nginx
        params = {
            "dn": request.META.get("HTTP_X_TDSK_SSL_CLIENT_S_DN"),
            "subject": request.META.get("HTTP_X_TDSK_SSL_CLIENT_S_DN"),
            "start_validity": request.META.get("HTTP_X_TDSK_SSL_CLIENT_VALIDITY_START"),
            "end_validity": request.META.get("HTTP_X_TDSK_SSL_CLIENT_VALIDITY_END"),
            "cert_verify": request.META.get("HTTP_X_TDSK_SSL_CLIENT_VERIFY"),
            "cert": request.META.get("HTTP_X_TDSK_SSL_CLIENT_CERT"),
            "ip": get_ip(request),
            "os": request.META.get("HTTP_USER_AGENT"),
        }

        params["is_valid"] = True if params['cert_verify'] == 'SUCCESS' else False
        params["date_start"] = parser.parse(params['start_validity'])
        params["date_end"] = parser.parse(params['end_validity'])

        return params
