import sys

from termidesk.audit.events import (WEB_ENTITY_ACTION, CLI_ENTITY_ACTION,
                                    POLICIES_GLOBAL_POLICY_CHANGED,
                                    POLICIES_GLOBAL_POLICY_DELETED,
                                    POLICIES_DEPLOYED_SERVICE_POLICY_CHANGED,
                                    POLICIES_DEPLOYED_SERVICE_POLICY_DELETED,
                                    USER_WORKPLACE_MESSAGE_SENT)
from uds.REST.handlers import NotFound
from uds.REST.model import OK, ModelHandler, OVERVIEW, TYPES, TABLEINFO, GUI

CREATE_ACTION = 'create'
READ_ACTION = 'read'
UPDATE_ACTION = 'update'
DELETE_ACTION = 'delete'


def set_crud_event(authenticator: str, username: str, ip: str,
                   entity: str, uuid: str, subtype: str,
                   name: str, action: str):
    event = WEB_ENTITY_ACTION(
        authenticator=authenticator,
        username=username,
        ip=ip,
        entity=entity,
        uuid=uuid,
        subtype=subtype,
        name=name,
        action=action
    ).generate_event()

    sys.audit(*event)


def set_crud_event_from_viewset(viewset, serializer, action):
    set_crud_event(
        viewset.request.user.manager.name,
        viewset.request.user.name,
        viewset.request.ip,
        viewset.model.__name__,
        serializer.data.get('id', ''),
        serializer.data.get('type') or serializer.data.get('data_type', ''),
        serializer.data.get('name', ''),
        action
    )


def raise_cli_crud_event(entity: str, uuid: str,
                         subtype: str, name: str, action: str, username: str = None) -> None:
    """Вызов события аудита при изменении объекта из командной строки"""
    if username is None:
        username = get_system_username()

    event = CLI_ENTITY_ACTION(
        username=username,
        entity=entity,
        uuid=uuid,
        subtype=subtype,
        name=name,
        action=action
    ).generate_event()

    sys.audit(*event)


def global_policies_update_event(username: str, authenticator_name: str, policy_name: str,
                                 value: str, authenticator_uuid: str, value_raw: str,) -> None:
    """Вызов события аудита при изменении глобальной политики"""

    event = POLICIES_GLOBAL_POLICY_CHANGED(
        username=username,
        authenticator_name=authenticator_name,
        policy_name=policy_name,
        value=value,
        authenticator_uuid=authenticator_uuid,
        value_raw=value_raw,
    ).generate_event()

    sys.audit(*event)


def global_policies_delete_event(username: str, authenticator_name: str, policy_name: str,
                                 authenticator_uuid: str) -> None:
    """Вызов события аудита при сбросе глобальной политики"""

    event = POLICIES_GLOBAL_POLICY_DELETED(
        username=username,
        authenticator_name=authenticator_name,
        policy_name=policy_name,
        authenticator_uuid=authenticator_uuid,
    ).generate_event()

    sys.audit(*event)


def policies_update_event(username: str, authenticator_name: str, policy_name: str,
                          deployed_service_name: str, value: str, authenticator_uuid: str,
                          deployed_service_uuid: str, value_raw: str,) -> None:
    """Вызов события аудита при изменении политики фонда"""

    event = POLICIES_DEPLOYED_SERVICE_POLICY_CHANGED(
        username=username,
        authenticator_name=authenticator_name,
        policy_name=policy_name,
        deployed_service_name=deployed_service_name,
        value=value,
        authenticator_uuid=authenticator_uuid,
        deployed_service_uuid=deployed_service_uuid,
        value_raw=value_raw,
    ).generate_event()

    sys.audit(*event)


def policies_delete_event(username: str, authenticator_name: str, policy_name: str,
                          deployed_service_name: str, authenticator_uuid: str,
                          deployed_service_uuid: str) -> None:
    """Вызов события аудита при удалении политики фонда"""

    event = POLICIES_DEPLOYED_SERVICE_POLICY_DELETED(
        username=username,
        authenticator_name=authenticator_name,
        policy_name=policy_name,
        deployed_service_name=deployed_service_name,
        authenticator_uuid=authenticator_uuid,
        deployed_service_uuid=deployed_service_uuid,
    ).generate_event()

    sys.audit(*event)


def user_workplace_message_sent_event(authenticator_name: str, authenticator_uuid: str,
                                      username: str, deployed_service_name: str,
                                      deployed_service_uuid: str, user_service_name: str,
                                      user_service_uuid: str, msg_level: str, msg_text: str) -> None:

    """Вызов события аудита при отправке сообщения в назначенные машины фонда"""

    event = USER_WORKPLACE_MESSAGE_SENT(
        authenticator_name=authenticator_name,
        authenticator_uuid=authenticator_uuid,
        username=username,
        deployed_service_name=deployed_service_name,
        deployed_service_uuid=deployed_service_uuid,
        user_service_name=user_service_name,
        user_service_uuid=user_service_uuid,
        msg_level=msg_level,
        msg_text=msg_text,
    ).generate_event()

    sys.audit(*event)


def get_system_username() -> str:
    """Получение имени системного пользователя"""
    import getpass
    system_user_name = getpass.getuser()
    return system_user_name


class AuditHandler(ModelHandler):
    """ Класс обработчик основных сущностей. """

    def create(self):
        res = super().put()

        if self.__is_detail():
            self.__set_update_parent()
            return res

        set_crud_event(
            self._user.manager.name,
            self._user.name,
            self._request.ip,
            self.model.__name__,
            res.get('id', ''),
            res.get('type', ''),
            res.get('name', ''),
            CREATE_ACTION
        )

        return res

    def put(self):

        if len(self._args) == 0:
            return self.create()
        else:
            return self.update()

    def update(self):
        res = super().put()

        if self.__is_detail():
            self.__set_update_parent()
            return res

        set_crud_event(
            self._user.manager.name,
            self._user.name,
            self._request.ip,
            self.model.__name__,
            res.get('id', ''),
            res.get('type', ''),
            res.get('name', ''),
            UPDATE_ACTION
        )

        return res

    def get(self):
        res = super().get()

        if len(self._args) == 0:
            return res

        value, *_ = self._args
        if len(self._args) == 1 and value not in [OVERVIEW, TYPES, TABLEINFO, GUI]:
            # Фиксация открытия интерфейса редактирования сущности
            set_crud_event(
                self._user.manager.name,
                self._user.name,
                self._request.ip,
                self.model.__name__,
                res.get('id', ''),
                res.get('type', ''),
                res.get('name', ''),
                READ_ACTION
            )

        return res

    def delete(self):
        try:
            item = self.model.objects.get(uuid=self._args[0].lower())
            subtype = item.data_type if hasattr(item, 'data_type') else ''
            name_object = item.name
        except self.model.DoesNotExist:
            raise NotFound('Element do not exists')

        res = super().delete()

        if self.__is_detail():
            self.__set_update_parent()
            return res

        if res == OK:
            set_crud_event(
                self._user.manager.name,
                self._user.name,
                self._request.ip,
                self.model.__name__,
                item.uuid,
                subtype,
                name_object,
                DELETE_ACTION
            )

        return res

    def __is_detail(self):
        """ Проверка является ли запрос действием над детальной. """
        return len(self._args) > 1

    def __set_update_parent(self):
        item = self.model.objects.filter(uuid=self._args[0])[0]
        subtype = item.data_type if hasattr(item, 'data_type') else ''

        set_crud_event(
            self._user.manager.name,
            self._user.name,
            self._request.ip,
            self.model.__name__,
            item.uuid,
            subtype,
            item.name,
            UPDATE_ACTION
        )
