"""Experimental realisation of SPICE protocol"""
# -*- coding: utf-8 -*-
#
# Copyright (c) 2012 Virtual Cable S.L.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# @author: Adolfo Gómez, dkmaster at dkmon dot com


# Modifications
# Copyright (c) 2017 Termilab, Dmitriy Dvoeglazov
# All rights reserved.
#
# Added proxy option
# @author: Dmitriy Dvoeglazov

from __future__ import unicode_literals

import logging

from django.utils.translation import gettext_noop as _

from termidesk.core.managers.PolicyManager import get_deployed_service_policy
from termidesk.core.util import wsptoken
from termidesk.policies.ClipboardEnabledPolicy import ClipboardEnabledPolicy
from termidesk.policies.ClientFileCopyEnabledPolicy import ClientFileCopyEnabledPolicy
from termidesk.policies.FullScreenPolicy import FullScreenPolicy
from termidesk.policies.ClientSmartCardRedirectEnabledPolicy import ClientSmartCardRedirectEnabledPolicy
from uds.core.util import OsDetector
from uds.core.util import tools
from uds.core.ui.UserInterface import gui
from uds.transports.SPICE.BaseSPICETransport import BaseSpiceTransport
from .RemoteViewerFile import RemoteViewerFile


import config

__updated__ = '2022-04-08'


logger = logging.getLogger(__name__)


class TDSKSPICETransport(BaseSpiceTransport):
    """
    Экспериментальный протокол на базе SPICE для поддержки нового клиента подключений `vdi-viewer`
    """
    typeName = _('SPICE (vdi-viewer, эксперим.)')
    typeType = 'TDSKSPICETransport'
    typeDescription = _('"Экспериментальный протокол на базе SPICE для нового клиента vdi-viewer')

    serverCertificate = BaseSpiceTransport.serverCertificate
    fullScreen = BaseSpiceTransport.fullScreen

    proxyUrl = gui.TextField(
        length=255,
        order=1,
        label=_('URL шлюза'),
        tooltip=_('Базовый URL вебсокет шлюза. (Формат: ws(s)://<хост>[:порт][/путь])'),
        tab=gui.ADVANCED_TAB
    )
    tunnelWait = gui.NumericField(
        length=3,
        label=_('Время ожидания соединения'),
        defvalue='20',
        minValue=1,
        maxValue=1800,
        order=2,
        tooltip=_('Максимальное время ожидание установления соединения'),
        required=True,
        tab=gui.TUNNEL_TAB
    )

    # Имя запускаемого клиента подключений
    client_filename = 'vdi-viewer'

    def getUDSTransportScript(self, user_service, transport, ip, os, user, password, request):
        user_service_instance = user_service.getInstance()

        con = user_service_instance.getConsoleConnection()

        logger.debug('Connection data: {}'.format(con))

        port = con.get('port', None)
        secure_port = con.get('secure_port', None)
        port = int(-1 if port is None else port)
        secure_port = int(-1 if secure_port is None else secure_port)

        with open(config.WSPROXY_PRIV_KEY_FILE) as f:
            sign_key = f.read()

        tun_url = ''
        tun_lifetime = self.tunnelWait.num()
        target_host = con['address']

        if self.proxyUrl.value:

            if port != -1:
                wstoken_data = {
                    'host': target_host,
                    'port': port,
                    'ssl_target': False,
                }
                wstoken = wsptoken.TokenEncoder(sign_key, lifetime=tun_lifetime).encode(wstoken_data)
                tun_url = '{}/{}'.format(self.proxyUrl.value, wstoken)

            if secure_port != -1:
                wsstoken_data = {
                    'host': target_host,
                    'port': secure_port,
                    'ssl_target': True,
                }
                wsstoken = wsptoken.TokenEncoder(sign_key, lifetime=tun_lifetime).encode(wsstoken_data)

                # Вебсокет шлюз поглощает TLS
                # На стороне клиента будет использоваться незашифрованный SPICE вариант
                # Шифрование, при необходимости, будет обеспечиваться HTTPS
                port = secure_port
                secure_port = -1
                tun_url = '{}/{}'.format(self.proxyUrl.value, wsstoken)

        # Убираем символ переноса каретки для корректного представления сертификата в скрипте
        ca = self.serverCertificate.value.replace('\r', '')

        clipboard_state = get_deployed_service_policy(ClipboardEnabledPolicy, user_service).get_value()
        filecopy_state = get_deployed_service_policy(ClientFileCopyEnabledPolicy, user_service).get_value()
        fullscreen_state = get_deployed_service_policy(FullScreenPolicy, user_service).get_value()
        pcsc_state = get_deployed_service_policy(
            ClientSmartCardRedirectEnabledPolicy, user_service).get_value()

        r = RemoteViewerFile(
            con['address'],
            port,
            secure_port,
            con['ticket']['value'],
            ca,
            con['cert_subject'],
            fullscreen=fullscreen_state,
            proxy=tun_url,
            termidesk_proxy=tun_url,
            termidesk_clipboard=clipboard_state,
            termidesk_filecopy=filecopy_state,
            termidesk_pcsc=pcsc_state,
        )

        m = tools.DictAsObj({
            'r': r,
            'client_filename': self.client_filename,
        })

        os = {
            OsDetector.Windows: 'windows',
            OsDetector.Linux: 'linux',
            OsDetector.Macintosh: 'macosx',
            OsDetector.FreeBSD: 'linux',
        }.get(os.OS)

        if os is None:
            return super().getUDSTransportScript(user_service, transport, ip, os, user, password, request)

        return self.getScript('scripts/{}/direct.py'.format(os)).format(m=m)


def get_clipboard_option_for_user_service(user_service):
    """Возвращает значение политики ClipboardEnabledPolicy для текущей машины"""

    value = ClipboardEnabledPolicy.value.defValue
    if PolicyManager().is_policies_enabled():
        policy = PolicyManager().get_deployed_service_policy(
            ClipboardEnabledPolicy,
            user_service.deployed_service,
        )
        value = policy.get_value()
    return value


def get_filecopy_option_for_user_service(user_service):
    """Возвращает значение политики ClientFileCopyEnabledPolicy для текущей машины"""

    value = ClientFileCopyEnabledPolicy.value.defValue
    if PolicyManager().is_policies_enabled():
        policy = PolicyManager().get_deployed_service_policy(
            ClientFileCopyEnabledPolicy,
            user_service.deployed_service,
        )
        value = policy.get_value()
    return value
