# -*- coding: utf-8 -*-

#
# Copyright (c) 2012 Virtual Cable S.L.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''

@author: Adolfo Gómez, dkmaster at dkmon dot com
'''

from __future__ import unicode_literals

import logging
import time
from concurrent.futures.thread import ThreadPoolExecutor
from itertools import filterfalse as exclude
from urllib.parse import urlparse

import ldap
import ldap.filter
import six
from ldap.controls.libldap import SimplePagedResultsControl

from django.utils.translation import gettext_noop as _

from termidesk.core.auths.ui.ExternalUsersUI import ExternalUsersUI
from uds.core.auths import Authenticator
from uds.core.auths.Exceptions import AuthenticatorException
from uds.core.ui.UserInterface import gui
from uds.core.util import log

__updated__ = '2016-04-18'

logger = logging.getLogger(__name__)

# Пул воркеров для отправки запросов к серверам ldap в многопоточном режиме.
# Решает проблему долгой аутентификации, когда подключение к одному или нескольким
# ldap серверам длится слишком долго, из-за чего gunicorn убивает процесс не дождавшись.
ldap_executor = ThreadPoolExecutor()

LDAP_RESULT_LIMIT = 50
LDAP_REFERRAL_BAN_SECONDS = 30


def _is_referral(record):
    # Функции поиска ldap возвращают кортеж из двух елементов - (dn, entry).
    # Параметр entry обычно является словарем.
    # См. https://www.python-ldap.org/en/python-ldap-3.3.0/reference/
    # ldap.html#ldap.LDAPObject.search_ext_s
    #
    # Однако, из MS AD могут иногда прилетать рефералы.
    # Для python-ldap записи могут выглядеть так:
    # (None, ['ldap://ForestDnsZones.testad.lan/DC=ForestDnsZones,DC=testad,DC=lan'])
    # Данный фильтр их распознает.
    dn, entry = record
    conditions = [
        dn is None,
        not isinstance(entry, dict)
    ]
    if all(conditions):
        return True

    # Запись не похожа на реферал
    return False


class ReferralEntry:
    """
    Объект реферальной ссылки LDAP сервера.

    Несет в себе имя сервера-реферала, а также время игнорирования этого сервера,
    если в предыдущий раз возникла ошибка подключения к этому серверу.

    При первой инициализации объект помещается в referral_registry.
    При последующих инициализациях возвращается ранее созданный объект из referral_registry.
    Таким образом осуществляется кэширование объектов в текущем процессе.

    За счет кэширования и информации о времени игнорирования удается освободить процессорное время,
    которое бы тратилось на бессмысленные запросы к недоступным серверам, на 50-60% в худшем случае:
    без кэширования авторизация длится 300-400 секунд, с кэшированием лишь 100-150, если использовать
    однопоточный режим.
    """
    referral_registry = {}

    banned_until = 0

    def __new__(cls, name):
        cls.referral_registry[name] = cls.referral_registry.get(name) or super().__new__(cls)
        return cls.referral_registry[name]

    def __init__(self, name):
        self.name = name

    def set_ban(self, seconds):
        """Установка времени бана в секундах от текущего процессорного времени"""
        self.banned_until = time.monotonic() + seconds

    @property
    def is_banned(self):
        return time.monotonic() - self.banned_until < 0


class SimpleLDAPAuthenticator(Authenticator):

    host = gui.TextField(length=64, label=_('Адрес Ldap'), order=1, tooltip=_('IP адрес или доменное имя Ldap сервера'), required=True)
    port = gui.NumericField(length=5, label=_('Порт'), defvalue='389', order=2, tooltip=_('Порт Ldap (обычно 389 для не-ssl и 636 для ssl)'), required=True)
    ssl = gui.CheckBoxField(label=_('Использовать SSL'), order=3, tooltip=_('Подключаться с помощью ssl на порт 636 вместо 389'))
    username = gui.TextField(length=512, label=_('Bind DN'), order=4, tooltip=_('Профиль аутентификации подключения — DN (Distinguished Name)'), required=True, tab=gui.CREDENTIALS_TAB)
    password = gui.PasswordField(lenth=32, label=_('Bind DN password'), order=5, tooltip=_('Пароль подключения'), required=True, tab=gui.CREDENTIALS_TAB)
    timeout = gui.NumericField(length=3, label=_('Timeout'), defvalue='10', order=6, tooltip=_('Максимальное время в секундах установления соединения'), required=True, minValue=1)
    ldapBase = gui.TextField(length=512, label=_('Base DN'), order=7, tooltip=_('Корень поиска пользователей и подразделений'), required=True, tab=_('Ldap info'))
    userClass = gui.TextField(length=64, label=_('User class name'), defvalue='posixAccount', order=8, tooltip=_('Имя класса пользователя (posixAccount/Person)'), required=True, tab=_('Ldap info'))
    userIdAttr = gui.TextField(length=64, label=_('User attr id'), defvalue='uid', order=9, tooltip=_('Атрибут уникального имени или идентификатора пользователя (uid/name)'), required=True, tab=_('Ldap info'))
    userNameAttr = gui.TextField(length=64, label=_('User attrs name'), defvalue='uid', order=10, tooltip=_('Список атрибутов через запятую, содержащий уникальные данные пользователя (uid, name)'), required=True, tab=_('Ldap info'))
    groupClass = gui.TextField(length=64, label=_('Group class name'), defvalue='posixGroup', order=11, tooltip=_('Имя класса группы (posixGroup/group)'), required=True, tab=_('Ldap info'))
    groupIdAttr = gui.TextField(length=64, label=_('Group attr name'), defvalue='cn', order=12, tooltip=_('Атрибут уникального имени группы (cn)'), required=True, tab=_('Ldap info'))
    memberAttr = gui.TextField(length=64, label=_('Group attr membership'), defvalue='memberUid', order=13, tooltip=_('Атрибут группы, содержащий принадлежащих ей пользователей (memberUid/member)'), required=True, tab=_('Ldap info'))

    typeName = _('MS Active Directory (LDAP)')
    typeType = 'SimpleLdapAuthenticator'
    typeDescription = _('Простая аутентификация через MS Active Directory (LDAP) по паролю')
    iconFile = 'auth.png'

    # If it has and external source where to get "new" users (groups must be declared inside UDS)
    isExternalSource = True
    # If we need to enter the password for this user
    needsPassword = False
    # Label for username field
    userNameLabel = _('Логин')
    userUI = ExternalUsersUI
    # Label for group field
    groupNameLabel = _("Группа")
    # Label for password field
    passwordLabel = _("Пароль")

    def __init__(self, dbAuth, environment, values=None):
        super(SimpleLDAPAuthenticator, self).__init__(dbAuth, environment, values)
        if values is not None:
            self._host = values['host']
            self._port = values['port']
            self._ssl = gui.strToBool(values['ssl'])
            self._username = values['username']
            self._password = values['password']
            self._timeout = values['timeout']
            self._ldapBase = values['ldapBase']
            self._userClass = values['userClass']
            self._groupClass = values['groupClass']
            self._userIdAttr = values['userIdAttr']
            self._groupIdAttr = values['groupIdAttr']
            self._memberAttr = values['memberAttr']
            self._userNameAttr = values['userNameAttr'].replace(' ', '')  # Removes white spaces
        else:
            self._host = None
            self._port = None
            self._ssl = None
            self._username = None
            self._password = None
            self._timeout = None
            self._ldapBase = None
            self._userClass = None
            self._groupClass = None
            self._userIdAttr = None
            self._groupIdAttr = None
            self._memberAttr = None
            self._userNameAttr = None
        self._connection = None

    def valuesDict(self):
        return {
            'host': self._host, 'port': self._port, 'ssl': gui.boolToStr(self._ssl),
            'username': self._username, 'password': self._password, 'timeout': self._timeout,
            'ldapBase': self._ldapBase, 'userClass': self._userClass, 'groupClass': self._groupClass,
            'userIdAttr': self._userIdAttr, 'groupIdAttr': self._groupIdAttr, 'memberAttr': self._memberAttr,
            'userNameAttr': self._userNameAttr
        }

    def __str__(self):
        return "Ldap Auth: {0}:{1}@{2}:{3}, base = {4}, userClass = {5}, groupClass = {6}, userIdAttr = {7}, groupIdAttr = {8}, memberAttr = {9}, userName attr = {10}".format(
            self._username, self._password, self._host, self._port, self._ldapBase, self._userClass, self._groupClass, self._userIdAttr, self._groupIdAttr, self._memberAttr,
            self._userNameAttr)

    def marshal(self):
        result =  '\t'.join(['v1',
                          self._host, self._port, gui.boolToStr(self._ssl), self._username, self._password, self._timeout,
                          self._ldapBase, self._userClass, self._groupClass, self._userIdAttr, self._groupIdAttr, self._memberAttr, self._userNameAttr])
        result = result.encode('utf8')
        return result

    def unmarshal(self, str_):
        data = str_.decode('utf8').split('\t')
        if data[0] == 'v1':
            logger.debug("Data: {0}".format(data[1:]))
            self._host, self._port, self._ssl, self._username, self._password, self._timeout, self._ldapBase, self._userClass, self._groupClass, self._userIdAttr, self._groupIdAttr, self._memberAttr, self._userNameAttr = data[1:]
            self._ssl = gui.strToBool(self._ssl)

    def __connection(self, username=None, password=None):
        if self._connection is None or username is not None:  # We want this method also to check credentials
            if six.PY2:
                if isinstance(username, six.text_type):
                    username = username.encode('utf8')
                if isinstance(password, six.text_type):
                    password = password.encode('utf8')
            l = None
            cache = False
            try:
                # ldap.set_option(ldap.OPT_DEBUG_LEVEL, 9)
                ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
                # Отключаем автоматическое следование за рефераллами как рекомендовано здесь:
                # https://www.python-ldap.org/en/python-ldap-3.3.0/faq.html#usage
                # Т.к это не работает с MS AD.
                ldap.set_option(ldap.OPT_REFERRALS, 0)

                schema = self._ssl and 'ldaps' or 'ldap'
                port = self._port != '389' and ':' + self._port or ''
                uri = "%s://%s%s" % (schema, self._host, port)
                logger.debug('Ldap uri: {0}'.format(uri))
                l = ldap.initialize(uri=uri)
                l.network_timeout = l.timeout = int(self._timeout)
                l.protocol_version = ldap.VERSION3

                if username is None:
                    cache = True
                    username = self._username
                    password = self._password

                l.simple_bind_s(who=username, cred=password)
            except ldap.LDAPError as e:
                str_ = _('Ldap connection error: ')
                # В подклассах LDAPError почему-то не всегда есть
                # атрибут message
                if not hasattr(e, 'message'):
                    str_ += str(e)
                    raise Exception(str_)

                if type(e.message) == dict:
                    str_ += 'info' in e.message and e.message['info'] + ',' or ''
                    str_ += 'desc' in e.message and e.message['desc'] or ''
                else:
                    str_ += str_(e)
                raise Exception(str_)
            if cache is True:
                self._connection = l
            else:
                return l  # Do not cache nor overwrite "global" connection
        return self._connection

    def _ref_connection(self, schema, host, port='389', username=None, password=None):
        '''
        Вспомогательный метод для получения подключения для реферальной ссылки
        '''
        if username is None:
            username = self._username
            password = self._password

        l = None
    
        try:
            # ldap.set_option(ldap.OPT_DEBUG_LEVEL, 9)
            ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
            # Отключаем автоматическое следование за рефераллами как рекомендовано здесь:
            # https://www.python-ldap.org/en/python-ldap-3.3.0/faq.html#usage
            # Т.к это не работает с MS AD.
            ldap.set_option(ldap.OPT_REFERRALS, 0)
    
            port_str = ':' + port if port!='389' else ''
            uri = "%s://%s%s" % (schema, host, port_str)
            logger.debug('Ldap uri: {0}'.format(uri))
            l = ldap.initialize(uri=uri)
            l.network_timeout = l.timeout = int(self._timeout)
            l.protocol_version = ldap.VERSION3
    
            l.simple_bind_s(who=username, cred=password)
        except ldap.LDAPError as e:
            str_ = _('Ldap connection error: ')
            if not hasattr(e, 'message'):
                str_ += str(e)
                raise Exception(str_)
                
            if type(e.message) == dict:
                str_ += 'info' in e.message and e.message['info'] + ',' or ''
                str_ += 'desc' in e.message and e.message['desc'] or ''
            else:
                str_ += str_(e)
            raise Exception(str_)
    
        return l

    def __getUser(self, username):
        """Try to get user.
        Return dict or None.
        All the exceptions will be suppressed.
        """
        try:
            return self.__getUserNoSuppress(username)
        except Exception as e:
            logger.exception('Exception: %s', e)
            return None

    def __getUserNoSuppress(self, username):
        """Try to get user.
        Return dict or raise exception if failed.
        """
        con = self.__connection()
        filter_ = '(&(objectClass={})({}={}))'.format(
            self._userClass,
            self._userIdAttr,
            ldap.filter.escape_filter_chars(username, 0),
        )
        attrlist = self._userNameAttr.split(',')
        attrlist += [self._userIdAttr, 'primaryGroupID', 'objectSid']
        logger.debug('Getuser filter_: {0}, attr list: {1}'.format(filter_, attrlist))

        res = self._ldap_search_paginated(
            connection=con,
            base=self._ldapBase,
            scope=ldap.SCOPE_SUBTREE,
            filterstr=filter_,
            attrlist=attrlist,
            timeout=int(self._timeout),
            sizelimit=LDAP_RESULT_LIMIT
        )

        if not res:
            raise Exception(
                'user not found: filter: {} attrlist: {} base: {}'.format(
                    filter_, attrlist, self._ldapBase,
                )
            )

        res = res[0]

        usr = dict((k, '') for k in attrlist)
        if six.PY3:
            # В python 3 из LDAP прилетают байтовые строки
            # Необходимо их преобразовать в unicode
            for attr in res[1]:
                vals = res[1][attr]
                decoded_vals = []
                for val in vals:
                    if attr == 'objectSid':
                        val = [hex(x) for x in val]
                    elif isinstance(val, six.binary_type):
                        val = val.decode('utf8')
                    decoded_vals.append(val)
                res[1][attr] = decoded_vals
                if self._userIdAttr.lower() == attr.lower():
                    # Названия атрибутов в MS AD/LDAP регистронезависимые.
                    # В ответе сервера атрибут может быть в одном регистре.
                    # В настройках атрибут id пользователя может быть в другом регистре.
                    # Форсируем наличие логина пользователя в атрибуте с
                    # указанным в настройках регистром для дальнейшего использования.
                    # TODO: Нужен рефакторинг с приведеним всех атрибутов из
                    # настроек и ответов сервера к нижнему регистру.
                    usr[self._userIdAttr] = decoded_vals
        usr.update(res[1])
        usr.update({'dn': res[0], '_id': username})
        logger.debug('Usr: {0}'.format(usr))
        return usr

    def __getGroup(self, groupName):
        try:
            con = self.__connection()
            filter_ = '(&(objectClass=%s)(%s=%s))' % \
                      (self._groupClass, self._groupIdAttr, ldap.filter.escape_filter_chars(groupName))
            memberAttrVal = self._memberAttr.encode('utf-8')
            if six.PY3:
                memberAttrVal = self._memberAttr
            attrlist = [memberAttrVal]
            logger.debug('Getgroup filter_: {0}, attr list {1}'.format(filter_, attrlist))
            res = self._ldap_search_paginated(
                connection=con,
                base=self._ldapBase,
                scope=ldap.SCOPE_SUBTREE,
                filterstr=filter_,
                attrlist=attrlist,
                timeout=int(self._timeout),
                sizelimit=LDAP_RESULT_LIMIT * 10
            )[0]

            grp = dict((k, ['']) for k in attrlist)
            grp.update(res[1])
            grp.update({'dn': res[0], '_id': groupName})
            logger.debug('Group: {0}'.format(grp))
            return grp
        except Exception as e:
            logger.exception('Exception: %s', e)
            return None

    def __getGroups(self, usr):
        """Возвращает список групп, к которым может принадлежать пользователь.

        Пользователь может быть привязан к группе явным образом - он находится в списке
        участников группы. Для поиска таких групп достаточно использовать фильтр
        "member=username" или "member=<distinguishedName>". Вместо "member" может использоваться
        любой другой атрибут, поэтому для гибкости, этот атрибут берется из поля "self._memberAttr".

        Пользователь может быть привязан к группе косвенным образом: группы пользователя могут
        представлять собой древовидную структуру. Поэтому, если пользователь принадлежит группе А,
        а группа А принадлежит группе Б, пользователь явно не входит в группу Б, то авторизация
        все равно должна проходить успешно, если в настройках аутентификатора указана только группа Б.
        Для поиска таких групп используется специальный фильтр "member:1.2.840.113556.1.4.1941",
        который позволяет производить поиск вверх по древовидной структуре и находить родительские группы.
        https://ldapwiki.com/wiki/Active%20Directory%20User%20Related%20Searches
        https://ldapwiki.com/wiki/1.2.840.113556.1.4.1941

        Дополнительно в список включается "первичная группа" (например, "Пользователи домена"),
        которая указана в параметре "primaryGroupID". Эту группу можно определить только по атрибуту
        "objectSid", который еще нужно сгенерировать на основе "primaryGroupID" и "objectSid" самого
        пользователя.
        """

        try:
            con = self.__connection()
            filter_string = '(&(objectClass={0})(|({1}={2})({1}={3})(member:1.2.840.113556.1.4.1941:={3})))'
            filter_string = filter_string.format(
                self._groupClass,
                self._memberAttr,
                ldap.filter.escape_filter_chars(usr['_id']),
                ldap.filter.escape_filter_chars(usr['dn']),
            )
            logger.debug('Filter: {0}'.format(filter_string))
            groupIdAttrVal = self._groupIdAttr.encode('utf8')
            if six.PY3:
                groupIdAttrVal = self._groupIdAttr

            res = self._ldap_search_paginated(
                connection=con,
                base=self._ldapBase,
                scope=ldap.SCOPE_SUBTREE,
                filterstr=filter_string,
                attrlist=[groupIdAttrVal],
                timeout=int(self._timeout),
                sizelimit=LDAP_RESULT_LIMIT * 10
            )

            primary_group = self.__getPrimaryGroup(usr)
            if primary_group:
                res.append(primary_group)

            groups = {}
            for g in res:
                _, entry = g
                # Нормализуем атрибуты ответа
                entry = {k.lower():v for k,v in entry.items()} 
                v = entry[self._groupIdAttr.lower()]
                if type(v) is not list:
                    v = [v]
                for gg in v:
                    if six.PY2:
                        groups[str(gg)] = g[0]
                    else:
                        if isinstance(gg, six.binary_type):
                            gg = gg.decode('utf8')
                        groups[gg] = g[0]
            logger.debug('Groups: {0}'.format(groups))
            return groups

        except Exception as e:
            logger.exception('Exception: %s', e)
            return {}

    def __getPrimaryGroup(self, usr):
        """Возвращает первичную группу пользователя.

        Если в параметрах пользователя присутствует primaryGroupID,
        то есть хотя бы одна первичная группа, например "Пользователи домена",
        которой принадлежит пользователь дополнительно.

        Группу нельзя искать по primaryGroupToken, но можно по objectSid.
        objectSid группы - это objectSid пользователя с замененными последними
        четырьмя байтами на primaryGroupID. Фильтр принимает байты в строковом
        представлении \00
        """

        con = self.__connection()
        result = None
        if usr.get('primaryGroupID') and usr.get('objectSid'):
            primary_group_id = [
                hex(x)
                for x in int(usr['primaryGroupID'][0]).to_bytes(4, byteorder='little')
            ]
            object_sid_str = ''.join(
                # Переводит '0x00' в '\00'
                r'\{x:0>2}'.format(x=x[2:])
                for x in usr['objectSid'][0][:-4] + primary_group_id
            )
            groups = self._ldap_search_paginated(
                connection=con,
                base=self._ldapBase,
                scope=ldap.SCOPE_SUBTREE,
                filterstr='(&(objectClass=group)(objectSid={}))'.format(object_sid_str),
                attrlist=[self._groupIdAttr],
                timeout=int(self._timeout),
                sizelimit=LDAP_RESULT_LIMIT * 10
            )
            result = next(iter(groups), None)
        return result

    def __getUserRealName(self, usr):
        '''
        Tries to extract the real name for this user. Will return all atttributes (joint)
        specified in _userNameAttr (comma separated).
        '''
        if six.PY3:
            name_attrs = [id_ for id_ in self._userNameAttr.split(',')]
            name_data = {k:usr[k] for k in name_attrs if usr.get(k)}
            # В метод могут передаваться как сырые данные из LDAP/AD, так и уже 
            # декодированные.
            # Поэтому необходимо нормализовать данные имени.
            for key in list(name_data.keys()):
                if isinstance(name_data[key], list):
                    new_list = []
                    for val in name_data[key]:
                        if isinstance(val, six.binary_type):
                            val = val.decode()
                        new_list.append(val)
                    name_data[key] = new_list
                else:
                    if isinstance(name_data[key], six.binary_type):
                        name_data[key] = name_data[key].decode()

            return ' '.join([(type(name_data.get(id_, '')) is list and ' '.join((k for k in name_data.get(id_, ''))) or (name_data.get(id_, ''))) for id_ in list(name_data.keys())]).strip()
        return ' '.join([(type(usr.get(id_, '')) is list and ' '.join((str(k) for k in usr.get(id_, ''))) or str(usr.get(id_, ''))) for id_ in self._userNameAttr.split(',')]).strip()
    
    def _ldap_search_paginated(self, connection, base, scope, filterstr=None, attrlist=None, attrsonly=0, serverctrls=None, clientctrls=None, timeout=-1, sizelimit=0, page_size=LDAP_RESULT_LIMIT):
        ''' 
            Поиск записей в LDAP с постраничным получением результатов.
            Может работать с большими количеством найденных записей.
            Возвращает результат в виде обычного списка.
        '''
        
        if serverctrls is None:
            serverctrls = []
        
        page_ctrl = SimplePagedResultsControl(criticality=True, size=page_size, cookie='')
        serverctrls.append(page_ctrl)
        res = []

        response = connection.search_ext(
            base=base, 
            scope=scope, 
            filterstr=filterstr, 
            attrlist=attrlist,
            attrsonly=attrsonly, 
            serverctrls=serverctrls,
            clientctrls=clientctrls,
            timeout=timeout,
            sizelimit=sizelimit
        )
        
        pages = 0
        while True:
            pages += 1
            rtype, rdata, rmsgid, recv_serverctrls = connection.result3(msgid=response)

            # Вначале добавляем обычные записи
            non_referrals = exclude(_is_referral, rdata)
            res.extend(non_referrals)

            # Пробуем обработать полученные реферальные ссылки
            referrals = filter(_is_referral, rdata)
            ref_res = self._retrieve_referrals(
                referrals=referrals,
                scope=scope,
                filterstr=filterstr,
                attrlist=attrlist,
                attrsonly=attrsonly, 
                clientctrls=clientctrls,
                timeout=timeout,
                sizelimit=sizelimit
            )
            res.extend(ref_res)

            controls = [control for control in recv_serverctrls
                        if control.controlType == SimplePagedResultsControl.controlType]
            if not controls:
                logger.warning('The server ignores RFC 2696 control')
                break
            if not controls[0].cookie:
                break
            page_ctrl.cookie = controls[0].cookie
            response = connection.search_ext(
                base=base, 
                scope=scope, 
                filterstr=filterstr, 
                attrlist=attrlist,
                attrsonly=attrsonly, 
                serverctrls=serverctrls,
                clientctrls=clientctrls,
                timeout=timeout,
                sizelimit=sizelimit
            )
        
        logger.debug("_ldap_search_paginated For base {} filter {} with sizelimit {} result pages: {} items: {}".format(base, filterstr, sizelimit, pages, len(res)))
        
        return res

    def _retrieve_referrals(self, referrals, scope, filterstr, attrlist, attrsonly,
                            clientctrls, timeout, sizelimit):
        '''
        Вспомогательный метод для загрузки данных по реферальным ссылкам.
        '''
        msg_tpl = ("_retrieve_referrals "
                   "Processing referal: %s")
        result = []

        def job(_record):
            _dn, _entry = _record
            url_parts = urlparse(_entry[0])

            ref_con = self._ref_connection(
                url_parts.scheme,
                url_parts.netloc,
                username=self._username,
                password=self._password
            )
            ref_base = url_parts.path[1:]

            return self._ldap_search_paginated(
                connection=ref_con,
                base=ref_base,
                scope=scope,
                filterstr=filterstr,
                attrlist=attrlist,
                attrsonly=attrsonly,
                clientctrls=clientctrls,
                timeout=timeout,
                sizelimit=sizelimit)

        futures = []
        for record in referrals:
            _dn, entry_list = record
            referral_entry = ReferralEntry(entry_list[0])

            if referral_entry.is_banned:
                continue

            future = ldap_executor.submit(job, record)
            futures.append((future, record))

        for future, record in futures:
            logger.debug(msg_tpl, record)
            _dn, entry_list = record
            referral_entry = ReferralEntry(entry_list[0])
            try:
                ref_res = future.result()
                result.extend(ref_res)
            except Exception as err:
                referral_entry.set_ban(LDAP_REFERRAL_BAN_SECONDS)
                if isinstance(err, ldap.LDAPError):
                    logger.exception('%s', err)
                else:
                    logger.exception('Unexpected error: %s', err)
                logger.error('Failed to connect to referal url: %s. Skipped.', entry_list)
                log_msg = (
                    'Подключения по реферальному адресу "{name}" заблокированы на '
                    '{seconds} секунд из-за ошибки "{error}"'
                ).format(
                    name=referral_entry.name,
                    seconds=LDAP_REFERRAL_BAN_SECONDS,
                    error=str(err),
                )
                log.doLog(self.dbAuthenticator(), log.ERROR, log_msg, log.INTERNAL)

        return result

    def transformUsername(self, username):
        usr = self.__getUser(username)
        if usr is None:
            raise Exception("Не удалось подключиться к MSAD")
        # Для нормализации логина пользователя используем соответствующее значение 
        # атрибута из настроек домена.
        return usr[self._userIdAttr][0]

    def authenticate(self, username, credentials, groupsManager):
        '''
        Must authenticate the user.
        We can have to different situations here:
           1.- The authenticator is external source, what means that users may be unknown to system before callig this
           2.- The authenticator isn't external source, what means that users have been manually added to system and are known before this call
        We receive the username, the credentials used (normally password, but can be a public key or something related to pk) and a group manager.
        The group manager is responsible for letting know the authenticator which groups we currently has active.
        @see: uds.core.auths.GroupsManager
        '''
        try:
            # Locate the user at LDAP
            usr = self.__getUserNoSuppress(username)

            if usr is None:
                return False

            # Let's see first if it credentials are fine
            self.__connection(usr['dn'], credentials)  # Will raise an exception if it can't connect

            groupsManager.validate(list(self.__getGroups(usr).keys()))

            return True

        except Exception as e:
            self._dbAuth.set_authenticator_error_reason(str(e))
            logger.exception(e)
            return False

    def getRealName(self, username):
        '''
        Tries to get the real name of an user
        '''
        res = self.__getUser(username)
        if res is None:
            return username
        return self.__getUserRealName(res)

    def createGroup(self, groupData):
        '''
        We must override this method in authenticators not based on external sources (i.e. database users, text file users, etc..)
        External sources already has its own groups and, at most, it can check if it exists on external source before accepting it
        Groups are only used in case of internal users (non external sources) that must know to witch groups this user belongs to
        @params groupData: a dict that has, at least, name, comments and active
        @return:  Raises an exception it things don't goes fine
        '''
        res = self.__getGroup(groupData['name'])
        if res is None:
            raise AuthenticatorException(_('Group not found'))

    def getGroups(self, username, groupsManager):
        '''
        Looks for the real groups to which the specified user belongs
        Updates groups manager with valid groups
        Remember to override it in derived authentication if needed (external auths will need this, for internal authenticators this is never used)
        '''
        user = self.__getUser(username)
        if user is None:
            raise AuthenticatorException(_('Username not found'))
        groupsManager.validate(list(self.__getGroups(user).keys()))

    def searchUsers(self, pattern):
        try:
            con = self.__connection()
            res = []
            
            records = self._ldap_search_paginated(
                connection=con,
                base=self._ldapBase, 
                scope=ldap.SCOPE_SUBTREE, 
                filterstr='(&(objectClass=%s)(%s=%s*))' % (self._userClass, self._userIdAttr, pattern),
                timeout=int(self._timeout),
                sizelimit=LDAP_RESULT_LIMIT,
            )
            for r in records:
                usrId = r[1].get(self._userIdAttr, '')
                usrId = type(usrId) == list and usrId[0] or usrId
                user = {
                    'id': usrId,
                    'name': self.__getUserRealName(r[1])
                }

                if six.PY3:
                    if isinstance(user['id'], six.binary_type):
                        user['id']=user['id'].decode('utf8')

                    if isinstance(user['name'], six.binary_type):
                        user['name']=user['name'].decode('utf8')

                res.append(user)
            return res
        except Exception as e:
            logger.exception("Exception: %s", e)
            raise AuthenticatorException(_('Too many results, be more specific'))

    def searchGroups(self, pattern):
        try:
            con = self.__connection()
            res = []
            records = self._ldap_search_paginated(
                connection=con,
                base=self._ldapBase, 
                scope=ldap.SCOPE_SUBTREE, 
                filterstr='(&(objectClass=%s)(%s=%s*))' % (self._groupClass, self._groupIdAttr, pattern),
                timeout=int(self._timeout),
                sizelimit=LDAP_RESULT_LIMIT,
            )
            
            for r in records:
                _, entry = r
                # Нормализуем атрибуты ответа
                entry = {k.lower():v for k,v in entry.items()} 
                grpId = entry.get(self._groupIdAttr.lower(), '')
                grpId = type(grpId) == list and grpId[0] or grpId
                res.append({
                    'id': grpId,
                    'name': grpId
                })
            return res
        except Exception as e:
            logger.exception("Exception: %s", e)
            raise AuthenticatorException(_('Too many results, be more specific'))

    @staticmethod
    def test(env, data):
        try:
            auth = SimpleLDAPAuthenticator(None, env, data)
            return auth.testConnection()
        except Exception as e:
            logger.exception("Exception found testing Simple LDAP auth {0}: {1}".format(e.__class__, e))
            return [False, "Error testing connection: {err}".format(err=str(e))]

    def testConnection(self):
        try:
            con = self.__connection()
        except Exception as e:
            return [False, str(e)]

        try:
            con.search_s(base=self._ldapBase, scope=ldap.SCOPE_BASE)
        except Exception:
            return [False, _('Ldap search base is incorrect')]

        try:
            if len(con.search_ext_s(base=self._ldapBase, scope=ldap.SCOPE_SUBTREE, filterstr='(objectClass=%s)' % self._userClass, sizelimit=1)) == 1:
                raise Exception()
            return [False, _('Ldap user class seems to be incorrect (no user found by that class)')]
        except Exception as e:
            # If found 1 or more, all right
            pass

        try:
            if len(con.search_ext_s(base=self._ldapBase, scope=ldap.SCOPE_SUBTREE, filterstr='(objectClass=%s)' % self._groupClass, sizelimit=1)) == 1:
                raise Exception()
            return [False, _('Ldap group class seems to be incorrect (no group found by that class)')]
        except Exception as e:
            # If found 1 or more, all right
            pass

        try:
            if len(con.search_ext_s(base=self._ldapBase, scope=ldap.SCOPE_SUBTREE, filterstr='(%s=*)' % self._userIdAttr, sizelimit=1)) == 1:
                raise Exception()
            return [False, _('Ldap user id attribute seems to be incorrect (no user found by that attribute)')]
        except Exception as e:
            # If found 1 or more, all right
            pass

        try:
            if len(con.search_ext_s(base=self._ldapBase, scope=ldap.SCOPE_SUBTREE, filterstr='(%s=*)' % self._groupIdAttr, sizelimit=1)) == 1:
                raise Exception()
            return [False, _('Ldap group id attribute seems to be incorrect (no group found by that attribute)')]
        except Exception as e:
            # If found 1 or more, all right
            pass

        # Now test objectclass and attribute of users
        try:
            if len(con.search_ext_s(base=self._ldapBase, scope=ldap.SCOPE_SUBTREE, filterstr='(&(objectClass=%s)(%s=*))' % (self._userClass, self._userIdAttr), sizelimit=1)) == 1:
                raise Exception()
            return [False, _('Ldap user class or user id attr is probably wrong (can\'t find any user with both conditions)')]
        except Exception as e:
            # If found 1 or more, all right
            pass

        # And group part, with membership
        try:
            memberAttrVal = self._memberAttr.encode('utf-8')
            if six.PY3:
                memberAttrVal = self._memberAttr
            
            res = self._ldap_search_paginated(
                connection=con,
                base=self._ldapBase, 
                scope=ldap.SCOPE_SUBTREE, 
                filterstr='(&(objectClass=%s)(%s=*))' % (self._groupClass, self._groupIdAttr), 
                attrlist=[memberAttrVal],
                timeout=int(self._timeout),
            )
            
            if len(res) == 0:
                raise Exception(_('Ldap group class or group id attr is probably wrong (can\'t find any group with both conditions)'))
            ok = False
            for r in res:
                if self._memberAttr in  r[1]:
                    ok = True
                    break
            if ok is False:
                raise Exception(_('Can\'t locate any group with the membership attribute specified'))
        except Exception as e:
            return [False, str(e)]

        return [True, _("Connection params seem correct, test was succesfully executed")]
