from __future__ import unicode_literals

import six

from uds.core import services
from uds.core.util.State import State
import logging
import json

logger = logging.getLogger(__name__)


class RDSDeployment(services.UserDeployment):
    """Interface for deployed MS RDS services.

    This class provides the needed logic for implementing an "consumable user service",
    that are the elements that the user will interact with.
    """
    def initialize(self):
        self._ip = self.service().getHostname()
        self._state = State.FINISHED
        self._reason = ''

    # Serializable needed methods
    def marshal(self):
        data = {
            'version': 1,
            'ip': self._ip,
            'state': self._state,
            'reason': self._reason,
        }

        return json.dumps(data)

    def unmarshal(self, str_):
        if isinstance(str_, six.binary_type):
            str_ = str_.decode('utf-8')
        data = json.loads(str_)
        if data['version'] == 1:
            self._ip = data.get('ip', '')
            self._reason = data['reason']
            self._state = data['state']
        else:
            raise Exception("Unsupported serialized data version {}".format(data['version']))

    def getIp(self):
        """Get an IP/FQDN of a MS RDS server from Provider"""
        return self._ip

    def setIp(self, ip):
        """This is an utility method, invoked by some os manager to notify what they thinks
        is the ip for this service. If you assign the service IP by your own methods, do
        not override this
        """
        logger.debug('Setting IP to %s (ignored)' % ip)

    def getUniqueId(self):
        """Returns the unique id == session id on MS RDS server"""
        return self.service().getUniqueId()

    def setReady(self):
        """The method is invoked whenever a machine is provided to an user, right
        before presenting it (via transport rendering) to the user.

        Excepted return values are State values RUNNING, FINISHED or ERROR.
        """
        self._logIP()
        self._state = State.FINISHED
        return self._state

    def deployForUser(self, user):
        """Deploys a user deployment as cache"""
        logger.debug("Starting deploy of {0} for user {1}".format(self._ip, user))
        self.dbservice().setInUse(True)
        self.dbservice().save()
        self._state = State.FINISHED
        return self._state

    def checkState(self):
        """Get the machine state from provider and convert it to expected state"""
        self._state = self.__checkMachineState()
        return self._state

    def reasonOfError(self):
        """Returns the reason of the error"""
        return self._reason

    def setupQueueForReplacement(self):
        pass

    def cancel(self):
        """Cancel represents a canceling of the current running operation, and
        can be invoked directly by an administration or by the clean up
        of the deployed service (indirectly).
        """
        return self.destroy()

    def destroy(self):
        """This method gives the oportunity to remove associated data (virtual machine, ...)
        for the user consumable this instance represents.

        Excepted return values are State values RUNNING, FINISHED or ERROR.
        """
        self._state = State.FINISHED
        return self._state

    def replace(self, data):
        # self._ip = ''
        # self._vmid = data.get('vm_id')
        pass

    def __checkMachineState(self):
        """Get a MS RDS current state

        Excepted return values are State values RUNNING, FINISHED or ERROR.
        """
        rds_state = self.service().getMachineState()
        if rds_state['is_rds']:
            state = State.FINISHED
        else:
            state = State.ERROR

        return state

    def _logIP(self):
        user_service = self.dbservice()
        user_service.logIP(self._ip)
