from __future__ import unicode_literals

from django.utils.translation import gettext_noop as _
from uds.core.services import ServiceProvider
from uds.core.ui import gui

from .RDSService import RDSService
from .rds_client import client

import logging


logger = logging.getLogger(__name__)

# Default unique_id value
DEFAULT_UNIQUE_ID = 'Unknown'


class Provider(ServiceProvider):
    """This class represents the MS RDS services provider

    In this class we provide:
       * The Provider functionality
       * The basic configuration parameters for the provider
       * The form fields needed by administrators to configure this provider
    """
    # : What kind of services we offer, this is a classes inherited from RDSService
    offers = [RDSService]
    # : Name to show the administrator
    typeName = _('Платформа MS RDS')
    # : Type used internally to identify this provider
    typeType = 'MSRDSPlatform'
    # : Description shown at administration interface for this provider
    typeDescription = _('Поставщик ресурсов на базе MS RDS')
    # : Icon file used as icon for this provider
    iconFile = 'provider.png'

    remoteHost = gui.TextField(oder=10,
                               length=64,
                               label=_('Адрес MS RDS'),
                               tooltip=_('IP-адрес или полное доменное имя (FQDN) MS RDS Server'),
                               required=True,
                               )
    remotePort = gui.TextField(oder=30,
                               length=64,
                               label=_('Порт RDS агента'),
                               tooltip=_('Порт MS RDS Server'),
                               defvalue='31000',
                               required=True,
                               )
    domain = gui.TextField(length=32,
                           label=_('Домен'),
                           order=4,
                           tooltip=_('Домен для подключения в MS RDS'),
                           required=True,
                           )
    username = gui.TextField(length=32,
                             label=_('Логин'),
                             order=5,
                             tooltip=_('Пользователь с необходимыми правами в MS RDS'),
                             required=True,
                             )
    password = gui.PasswordField(lenth=32,
                                 label=_('Пароль'),
                                 order=6,
                                 tooltip=_('Пароль пользователя MS RDS'),
                                 required=True,
                                 )

    _api = None

    def __getApi(self):
        """Returns the connection API object for RDS"""
        if self._api is None:
            self._api = client.RDSClient(self.remoteHost.value, self.remotePort.value)
        return self._api

    def getMachineState(self):
        """Returns the state of the MS RDS server"""
        return self.__getApi().check_rds_instance()

    def getUniqueId(self):
        """Default unique_id value"""
        user_session = DEFAULT_UNIQUE_ID
        return user_session

    def get_sessions(self):
        rds_sessions = self.__getApi().get_sessions()
        return rds_sessions

    def testConnection(self):
        """Test that conection to MS RDS server is fine"""
        test = self.__getApi().test_connection(self.username.value, self.password.value, self.domain.value)
        return test

    @staticmethod
    def test(env, data):
        """Test MS RDS connectivity

        Args:
            env: environment passed for testing (temporal environment passed)
            data: data passed for testing (data obtained from the form
            definition)

        Returns:
            Array of two elements, first is True of False, depending on test
            (True is all right, false is error),
            second is a string with error.
        """
        ov = Provider(env, data)
        test = ov.testConnection()
        status, message = test
        test_result = [status, _(message)]
        return test_result
