import os

from server.celery import app as celery
from uds.core.jobs import Job
from uds.models import ServicePool, getSqlDatetime
from uds.core.util.State import State

from .RDSService import RDSService
from .RDSProvider import DEFAULT_UNIQUE_ID


# Default task frequency
ENV_TIMEOUT_VAR = 'TDSK_ID_SESSION_CHECK_TIMEOUT'
DEFAULT_TIMEOUT = 2
# Default value of missing unique_id
MISSING_UNIQUE_ID = DEFAULT_UNIQUE_ID.upper()


class RDSSessionUpdate(celery.Task, Job):
    """Periodic task.
    Updates the pools sessions.
    If the session is valid on the RDS server: actualize uniqueID = sessionID.
    If session is not available on RDS server: set pool session status = REMOVED.

    """
    # Historical spelling error, should be spelled like this: "frecuency".
    frecuency = os.getenv(ENV_TIMEOUT_VAR, DEFAULT_TIMEOUT)
    friendly_name = 'RDS Sessions Updater'
    celery_name = 'J.uds.core.workers.Jobs.RDSSessionUpdate'
    name = celery_name
    celery_compatible = True
    rds_sessions = {}

    def __call__(self, *args, **kwargs):
        return self.run()

    def get_rds_sessions(self):
        # Get all active RDS pools.
        service_pools = ServicePool.objects.filter(service__data_type=RDSService.typeType,
                                                   state=State.ACTIVE)
        for pool in service_pools:
            # Get remote RDS server ip and port via pool Provider.
            provider = pool.service.provider.getInstance()
            # Get active RDS sessions in format {user: sessionID} for current pool.
            pool_sessions = provider.get_sessions().get('sessions')
            # Update self.rds_sessions in format: {pool_db_object: {user1: session1, user2: session2, ...}}.
            self.rds_sessions[pool] = pool_sessions

    def deactivate_sessions(self):
        for pool, session_ids in self.rds_sessions.items():
            # Get active pool sessions with unique_id.
            # If service user not in rds_sessions: update state = REMOVED.
            pool.userServices\
                .filter(state=State.USABLE).filter(in_use=True)\
                .exclude(unique_id__in=session_ids.values()).exclude(unique_id=MISSING_UNIQUE_ID)\
                .update(state=State.REMOVED, state_date=getSqlDatetime())

    def activate_sessions(self):
        for pool, session_ids in self.rds_sessions.items():
            # Get active pool sessions with missing unique_id.
            user_services_to_update = pool.userServices\
                .filter(state=State.USABLE).filter(unique_id=MISSING_UNIQUE_ID)
            # Iterate over pool sessions to validate them.
            for user_service in user_services_to_update:
                try:
                    rds_session_id = str(session_ids[user_service.user.name])
                    user_service.unique_id = rds_session_id
                    user_service.friendly_name = rds_session_id
                    user_service.save()
                except KeyError:
                    continue

    def run(self):
        # Get all active rds sessions.
        self.get_rds_sessions()
        # Updates state to "REMOVED" for inactive sessions
        self.deactivate_sessions()
        # Updates `unique_id` and `friendly_name` for active RDS sessions,
        # when user has an active rdp connection.
        self.activate_sessions()


celery.tasks.register(RDSSessionUpdate)
