from http import HTTPStatus
import json
import socket

import requests


# RDS-agent URL
AGENT_HOST = '127.0.0.1:8000'
# Agent connection error api_response
AGENT_CONNECTION_ERROR = {'status': HTTPStatus.NOT_FOUND.value}
# Agent available status_code
AGENT_AVAILABLE = 'ok'
# Timeout before request expires
DEFAULT_TIMEOUT = 10
# RDS-agent default port
DEFAULT_PORT = 31000

# Variable keys for status messages
SUCCESS = 'success'
HOSTNAME_UNRESOLVED = 'hostname_unresolved'
LOGIN_FAILED = 'login_failed'
CONNECTION_ERROR = 'connection_error'

# Status messages for provider test
STATUS_MESSAGES = {
    SUCCESS: (True, 'Тест успешно пройден'),
    HOSTNAME_UNRESOLVED: (False, 'Не удалось определить имя сервера, проверьте имя сервера RDS'),
    LOGIN_FAILED: (False, 'Неверное имя пользователя или пароль.'),
    CONNECTION_ERROR: (False, 'Не удалось подключиться к сессионному агенту. Проверьте, что сессионный '
                              'агент установлен и запущен.')
}


def send_api_response(func):
    """Methods decorator.
    Sending json-response based on api-response.
    Checking if api_response is not None:
        success: checking if HTTP-status code == 200:
            success: return json_response;
            failure: return actual status_code;
        failure: return status_code == HTTPStatus.NOT_FOUND.

    Arguments:
        func: function. Returns response from api-request.

    """

    def wrapper(*args, **kwargs) -> dict:
        api_response = func(*args, **kwargs)
        if api_response:
            if api_response.status_code == HTTPStatus.OK:
                json_response = api_response.json()
            else:
                json_response = {'status': api_response.status_code}
            return json_response
        else:
            return AGENT_CONNECTION_ERROR

    return wrapper


# noinspection PyArgumentList
class RDSClient:
    """RDS client for data exchange with RDS agent"""
    def __init__(self, ip: str = None, port: str = None) -> None:
        self._ip = ip
        self._port = port if port else DEFAULT_PORT
        self._host = '{ip}:{port}'.format(ip=self._ip, port=self._port) if ip else AGENT_HOST
        self._timeout = DEFAULT_TIMEOUT

    def __get_api_uri(self, destination: str) -> str:
        """Form URI for API-request based on destination.

        Arguments:
            destination: API-destination.

        """
        api_uri = 'http://{host}/{destination}'.format(host=self._host, destination=destination)
        return api_uri

    # noinspection PyMethodParameters
    def check_agent_availability(func):
        """Methods decorator.
        Checking agent availability based on get api-request /public/status.
        Must return {'status_code': 'ok'}:
            success: return api_response based on request in method.
            failure: return api request = None.

        Arguments:
            func: function. Returns response from api-request.

        """

        # noinspection PyCallingNonCallable
        def wrapper(self, *args, **kwargs):
            api_response = None
            try:
                agent_available = requests.get(self.__get_api_uri('public/status'), timeout=DEFAULT_TIMEOUT)
                if agent_available.json()['status'] == AGENT_AVAILABLE:
                    api_response = func(self, *args, **kwargs)
            except (json.JSONDecodeError, requests.ConnectionError):
                pass
            return api_response

        return wrapper

    @send_api_response
    @check_agent_availability
    def check_rds_instance(self) -> requests.Response:
        """Check RDS instance"""
        destination = 'rds/test'
        api_response = requests.get(self.__get_api_uri(destination), timeout=self._timeout)
        return api_response

    @send_api_response
    @check_agent_availability
    def authorize(self, login: str, password: str) -> requests.Response:
        """Authorization in RDS.

        Arguments:
            login: user login,
            password: user password.

        """
        content = {'login': login, 'password': password}
        # Checking if login and password is string. If not: add them to list.
        wrong_args_types = [arg[0] for arg in content.items() if not isinstance(arg[1], str)]
        # Raise TypeError if list is not empty. Show what arg(s) cause problem.
        if wrong_args_types:
            raise TypeError('Arguments must be string-type: {}.'.format(', '.join(wrong_args_types)))
        destination = 'rds/login'
        api_response = requests.post(self.__get_api_uri(destination), json=content, timeout=self._timeout)
        return api_response

    @send_api_response
    @check_agent_availability
    def get_sessions(self):
        """Get a list of RDS sessions"""
        destination = 'rds/sessions'
        api_response = requests.get(self.__get_api_uri(destination), timeout=self._timeout)
        return api_response

    def _check_domain_resolving(self) -> None:
        """Checking if domain name is resolvable and port is available.
        One of the resource provider checks:
            success: returns nothing;
            failure: raise exception.
        """
        error = None
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(5)
        try:
            s.connect((self._ip, int(self._port)))
            s.shutdown(socket.SHUT_RDWR)
        except Exception as ex:
            error = ex
        finally:
            s.close()
            if error:
                raise error

    def _test_authorization(self, login: str, password: str, domain: str) -> (None, str):
        """Test authorization with credentials from the provider.
        One of the resource provider checks:
            success: returns error_code;
            failure: raise (requests.ConnectTimeout, requests.ConnectionError).
        """
        destination = 'rds/testconnect'
        content = {'login': login, 'password': password, 'domain': domain}
        api_response = requests.post(self.__get_api_uri(destination),
                                     json=content,
                                     timeout=self._timeout).json()
        error_code = api_response.get('error')
        return error_code

    def test_connection(self, login: str, password: str, domain: str = None) -> tuple:
        """Checking the availability of a resource provider

        Arguments:
            login: RDS admin login,
            password: RDS admin password,
            domain: RDS domain (if needed).

        """
        try:
            self._check_domain_resolving()
            login_unsuccessful = self._test_authorization(login, password, domain)
            if login_unsuccessful:
                raise requests.ReadTimeout
            else:
                test_result = STATUS_MESSAGES[SUCCESS]

        except socket.gaierror:
            test_result = STATUS_MESSAGES[HOSTNAME_UNRESOLVED]
        except (ConnectionRefusedError, socket.timeout, requests.ConnectTimeout, requests.ConnectionError):
            test_result = STATUS_MESSAGES[CONNECTION_ERROR]
        except requests.ReadTimeout:
            test_result = STATUS_MESSAGES[LOGIN_FAILED]

        return test_result
